import numpy as np
from math import ceil
from hapla import shared_cy
from hapla import admix_cy

##### hapla - functions #####
### hapla struct
# SVD through eigendecomposition
def eigSVD(H):
	D, V = np.linalg.eigh(np.dot(H.T, H))
	S = np.sqrt(D)
	U = np.dot(H, V*(1.0/S))
	return np.ascontiguousarray(U[:,::-1]), np.ascontiguousarray(S[::-1]), \
		np.ascontiguousarray(V[:,::-1])

# Randomized PCA with dynamic shift
def randomizedSVD(Z_agg, p_vec, a_vec, K, batch, power, rng):
	M, N = Z_agg.shape
	W = ceil(M/batch)
	a = 0.0
	L = K + 10
	H = np.zeros((N, L))
	X = np.zeros((batch, N))
	A = rng.standard_normal(size=(M, L))

	# Prime iteration
	for w in np.arange(W):
		M_w = w*batch
		if w == (W-1): # Last batch
			X = np.zeros((M - M_w, N))
		shared_cy.batchZ(Z_agg, X, p_vec, a_vec, M_w)
		H += np.dot(X.T, A[M_w:(M_w + X.shape[0])])
	Q, _, _ = eigSVD(H)
	H.fill(0.0)

	# Power iterations
	for p in np.arange(power):
		print(f"\rPower iteration {p+1}/{power}", end="")
		X = np.zeros((batch, N))
		for w in np.arange(W):
			M_w = w*batch
			if w == (W-1): # Last batch
				X = np.zeros((M - M_w, N))
			shared_cy.batchZ(Z_agg, X, p_vec, a_vec, M_w)
			A[M_w:(M_w + X.shape[0])] = np.dot(X, Q)
			H += np.dot(X.T, A[M_w:(M_w + X.shape[0])])
		H -= a*Q
		Q, S, _ = eigSVD(H)
		H.fill(0.0)
		if S[-1] > a:
			a = 0.5*(S[-1] + a)

	# Extract singular vectors
	X = np.zeros((batch, N))
	for w in np.arange(W):
		M_w = w*batch
		if w == (W-1): # Last batch
			X = np.zeros((M - M_w, N))
		shared_cy.batchZ(Z_agg, X, p_vec, a_vec, M_w)
		A[M_w:(M_w + X.shape[0])] = np.dot(X, Q)
	U, S, V = eigSVD(A)
	return U[:,:K], S[:K], np.dot(Q, V)[:,:K]



### hapla admix
# Update for ancestry estimation
def steps(Z, P, Q, P_tmp, Q_tmp, k_vec, c_vec, y):
	admix_cy.updateP(Z, P, Q, P_tmp, Q_tmp, k_vec, c_vec)
	admix_cy.updateQ(Q, Q_tmp, Z.shape[0])
	if y is not None:
		admix_cy.superQ(Q, y)

# Accelerated update for ancestry estimation
def quasi(Z, P0, Q0, P_tmp, Q_tmp, P1, P2, Q1, Q2, k_vec, c_vec, y):
	# 1st EM step
	admix_cy.accelP(Z, P0, P1, Q0, P_tmp, Q_tmp, k_vec, c_vec)
	admix_cy.accelQ(Q0, Q1, Q_tmp, Z.shape[0])
	if y is not None:
		admix_cy.superQ(Q1, y)

	# 2nd EM step
	admix_cy.accelP(Z, P1, P2, Q1, P_tmp, Q_tmp, k_vec, c_vec)
	admix_cy.accelQ(Q1, Q2, Q_tmp, Z.shape[0])
	if y is not None:
		admix_cy.superQ(Q2, y)

	# Acceleation update
	admix_cy.alphaP(P0, P1, P2, k_vec, c_vec, Q0.shape[1])
	admix_cy.alphaQ(Q0, Q1, Q2)
	if y is not None:
		admix_cy.superQ(Q0, y)



# cython: language_level=3, boundscheck=False, wraparound=False, initializedcheck=False, cdivision=True
cimport numpy as np
cimport openmp as omp
from cython.parallel import parallel, prange
from libc.math cimport fmin, fmax, log
from libc.stdint cimport uint8_t, uint32_t
from libc.stdlib cimport calloc, free

cdef double PRO_MIN = 1e-5
cdef double PRO_MAX = 1.0-(1e-5)
cdef double ACC_MIN = 1.0
cdef double ACC_MAX = 256.0

##### hapla - ancestry estimation #####
# Truncate parameters to domain
cdef double _project(
		const double s
	) noexcept nogil:
	return fmin(fmax(s, PRO_MIN), PRO_MAX)

# Estimate inverse individual allele frequency
cdef double _computeH(
		const double* p, const double* q, const size_t z, const size_t K, const size_t B
	) noexcept nogil:
	cdef:
		double h = 0.0
		size_t k
	for k in range(K):
		h += p[k*B + z]*q[k]
	return 1.0/h

# Estimate individual allele frequency
cdef double _computeL(
		const double* p, const double* q, const size_t z, const size_t K, const size_t B
	) noexcept nogil:
	cdef:
		double h = 0.0
		size_t k
	for k in range(K):
		h += p[k*B + z]*q[k]
	return log(h)

# Inner loop updates for temp P and Q
cdef void _inner(
		const double* p, const double* q, double* p_tmp, double* q_thr, const double h, const size_t z, 
		const size_t K, const size_t B
	) noexcept nogil:
	cdef size_t k
	for k in range(K):
		p_tmp[k*B + z] += q[k]*h
		q_thr[k] += p[k*B + z]*h

# Outer loop update for P
cdef void _outerP(
		double* p, double* p_tmp, const size_t K, const size_t B
	) noexcept nogil:
	cdef:
		double sumA, sumB
		size_t c, k, s
	for k in range(K):
		s = k*B
		sumA = 0.0
		sumB = 0.0
		for c in range(B):
			p[s+c] = p_tmp[s+c]*p[s+c]
			sumA += p[s+c]
		for c in range(B):
			p[s+c] = _project(p[s+c]/sumA)
			sumB += p[s+c]
		for c in range(B):
			p[s+c] /= sumB
			p_tmp[s+c] = 0.0

# Outer loop accelerated update for P
cdef void _outerAccelP(
		const double* p, double* p_new, double* p_tmp, const size_t K, const size_t B
	) noexcept nogil:
	cdef:
		double sumA, sumB
		size_t c, k, s
	for k in range(K):
		s = k*B
		sumA = 0.0
		sumB = 0.0
		for c in range(B):
			p_new[s+c] = p_tmp[s+c]*p[s+c]
			sumA += p_new[s+c]
		for c in range(B):
			p_new[s+c] = _project(p_new[s+c]/sumA)
			sumB += p_new[s+c]
		for c in range(B):
			p_new[s+c] /= sumB
			p_tmp[s+c] = 0.0

# Outer loop update for Q
cdef void _outerQ(
		double* q, double* q_tmp, const double S, const size_t K
	) noexcept nogil:
	cdef:
		double sumQ = 0.0
		size_t k
	for k in range(K):
		q[k] = _project(q[k]*q_tmp[k]*S)
		sumQ += q[k]
	for k in range(K):
		q[k] /= sumQ
		q_tmp[k] = 0.0

# Outer loop accelerated update for Q
cdef void _outerAccelQ(
		const double* q, double* q_new, double* q_tmp, const double S, const size_t K
	) noexcept nogil:
	cdef:
		double sumQ = 0.0
		size_t k
	for k in range(K):
		q_new[k] = _project(q[k]*q_tmp[k]*S)
		sumQ += q_new[k]
	for k in range(K):
		q_new[k] /= sumQ
		q_tmp[k] = 0.0

# Estimate QN factor
cdef double _factorAccel(
		const double* v0, const double* v1, const double* v2, const size_t I
	) noexcept nogil:
	cdef:
		double sum1 = 0.0
		double sum2 = 0.0
		double u, v
		size_t i
	for i in prange(I):
		u = v1[i] - v0[i]
		v = v2[i] - v1[i] - u
		sum1 += u*u
		sum2 += u*v
	return fmin(fmax(-(sum1/sum2), ACC_MIN), ACC_MAX)


### Update functions
# Create P matrix from array
cpdef void createP(
		double[::1] P, const uint8_t[::1] k_vec, const uint32_t[::1] c_vec, const size_t K
	) noexcept nogil:
	cdef:
		double sumP
		size_t W = k_vec.shape[0]
		size_t c, k, l, s, w, B
	for w in prange(W):
		B = <size_t>k_vec[w]
		l = <size_t>c_vec[w]
		for k in range(K):
			s = l + k*B
			sumP = 0.0
			for c in range(B):
				sumP = sumP + P[s+c]
			for c in range(B):
				P[s+c] /= sumP

# Update Q in supervised mode
cpdef void superQ(
		double[:,::1] Q, const uint8_t[::1] y
	) noexcept nogil:
	cdef:
		double sumQ
		size_t K = Q.shape[1]
		size_t N = Q.shape[0]
		size_t i, k
	for i in prange(N):
		if y[i] > 0:
			sumQ = 0.0
			for k in range(K):
				if k == (y[i] - 1):
					Q[i,k] = PRO_MAX
				else:
					Q[i,k] = PRO_MIN
				sumQ = sumQ + Q[i,k]
			for k in range(K):
				Q[i,k] /= sumQ

# Update P and Q temp arrays
cpdef void updateP(
		const uint8_t[:,::1] Z, double[::1] P, const double[:,::1] Q, double[::1] P_tmp, double[:,::1] Q_tmp, 
		const uint8_t[::1] k_vec, const uint32_t[::1] c_vec
	) noexcept nogil:
	cdef:
		double h
		double* pl
		double* pt
		double* Q_thr
		omp.omp_lock_t mutex
		size_t W = Z.shape[0]
		size_t N = Z.shape[1]
		size_t K = Q.shape[1]
		size_t z, B
		size_t i, l, n, w, x, y
	omp.omp_init_lock(&mutex)
	with nogil, parallel():
		Q_thr = <double*>calloc((N//2)*K, sizeof(double))
		for w in prange(W):
			B = <size_t>k_vec[w]
			l = <size_t>c_vec[w]
			pl = &P[l]
			pt = &P_tmp[l]
			for i in range(N):
				n = i//2
				z = <size_t>Z[w,i]
				h = _computeH(pl, &Q[n,0], z, K, B)
				_inner(pl, &Q[n,0], pt, &Q_thr[n*K], h, z, K, B)
			_outerP(pl, pt, K, B)

		# omp critical
		omp.omp_set_lock(&mutex)
		for x in range(N//2):
			for y in range(K):
				Q_tmp[x,y] += Q_thr[x*K + y]
		omp.omp_unset_lock(&mutex)
		free(Q_thr)
	omp.omp_destroy_lock(&mutex)

# Accelerated update P and Q temp arrays
cpdef void accelP(
		const uint8_t[:,::1] Z, double[::1] P, double[::1] P_new, const double[:,::1] Q, double[::1] P_tmp, 
		double[:,::1] Q_tmp, const uint8_t[::1] k_vec, const uint32_t[::1] c_vec
	) noexcept nogil:
	cdef:
		double h
		double* pl
		double* pt
		double* Q_thr
		omp.omp_lock_t mutex
		size_t W = Z.shape[0]
		size_t N = Z.shape[1]
		size_t K = Q.shape[1]
		size_t z, B
		size_t i, l, n, w, x, y
	omp.omp_init_lock(&mutex)
	with nogil, parallel():
		Q_thr = <double*>calloc((N//2)*K, sizeof(double))
		for w in prange(W):
			B = <size_t>k_vec[w]
			l = <size_t>c_vec[w]
			pl = &P[l]
			pt = &P_tmp[l]
			for i in range(N):
				n = i//2
				z = <size_t>Z[w,i]
				h = _computeH(pl, &Q[n,0], z, K, B)
				_inner(pl, &Q[n,0], pt, &Q_thr[n*K], h, z, K, B)
			_outerAccelP(pl, &P_new[l], pt, K, B)
		
		# omp critical
		omp.omp_set_lock(&mutex)
		for x in range(N//2):
			for y in range(K):
				Q_tmp[x,y] += Q_thr[x*K + y]
		omp.omp_unset_lock(&mutex)
		free(Q_thr)
	omp.omp_destroy_lock(&mutex)

# Accelerated jump for P (QN)
cpdef void alphaP(
		double[::1] P0, const double[::1] P1, const double[::1] P2, const uint8_t[::1] k_vec, 
		const uint32_t[::1] c_vec, const size_t K
	) noexcept nogil:
	cdef:
		double c1, c2, sumP
		size_t M = P0.shape[0]
		size_t W = k_vec.shape[0]
		size_t c, k, l, s, w, B
	c1 = _factorAccel(&P0[0], &P1[0], &P2[0], M)
	c2 = 1.0 - c1
	for w in prange(W):
		B = <size_t>k_vec[w]
		l = <size_t>c_vec[w]
		for k in range(K):
			s = l + k*B
			sumP = 0.0
			for c in range(B):
				P0[s+c] = _project(c2*P1[s+c] + c1*P2[s+c])
				sumP = sumP + P0[s+c]
			for c in range(B):
				P0[s+c] /= sumP

# Update Q
cpdef void updateQ(
		double[:,::1] Q, double[:,::1] Q_tmp, const size_t W
	) noexcept nogil:
	cdef:
		double S = 1.0/<double>(2*W)
		size_t N = Q.shape[0]
		size_t K = Q.shape[1]
		size_t i, k
	for i in prange(N):
		_outerQ(&Q[i,0], &Q_tmp[i,0], S, K)

# Accelerated update Q
cpdef void accelQ(
		const double[:,::1] Q, double[:,::1] Q_new, double[:,::1] Q_tmp, const size_t W
	) noexcept nogil:
	cdef:
		double S = 1.0/<double>(2*W)
		size_t N = Q.shape[0]
		size_t K = Q.shape[1]
		size_t i, k
	for i in prange(N):
		_outerAccelQ(&Q[i,0], &Q_new[i,0], &Q_tmp[i,0], S, K)

# Accelerated jump for Q (QN)
cpdef void alphaQ(
		double[:,::1] Q0, const double[:,::1] Q1, const double[:,::1] Q2
	) noexcept nogil:
	cdef:
		double c1, c2, sumQ
		size_t N = Q0.shape[0]
		size_t K = Q0.shape[1]
		size_t i, k
	c1 = _factorAccel(&Q0[0,0], &Q1[0,0], &Q2[0,0], N*K)
	c2 = 1.0 - c1
	for i in prange(N):
		sumQ = 0.0
		for k in range(K):
			Q0[i,k] = _project(c2*Q1[i,k] + c1*Q2[i,k])
			sumQ = sumQ + Q0[i,k]
		for k in range(K):
			Q0[i,k] /= sumQ	

# Log-likelihood
cpdef double loglike(
		const uint8_t[:,::1] Z, double[::1] P, const double[:,::1] Q, const uint8_t[::1] k_vec, 
		const uint32_t[::1] c_vec
	) noexcept nogil:
	cdef:
		double res = 0.0
		double h
		double* pl
		size_t W = Z.shape[0]
		size_t N = Z.shape[1]
		size_t K = Q.shape[1]
		size_t B
		size_t i, w
	for w in prange(W):
		B = <size_t>k_vec[w]
		pl = &P[c_vec[w]]
		for i in range(N):
			res += _computeL(pl, &Q[i//2,0], <size_t>Z[w,i], K, B)
	return res



"""
hapla.
Ancestry estimation using fastmixture mini-batch EM algorithm.
"""

__author__ = "Jonas Meisner"

# Libraries
import os
from time import time

##### hapla admix #####
def main(args):
	print("-----------------------------------")
	print("hapla by Jonas Meisner (v0.23.0)")
	print(f"hapla admix using {args.threads} thread(s)")
	print("-----------------------------------\n")

	# Check input
	assert (args.filelist is not None) or (args.clusters is not None), "No input data (--filelist or --clusters)!"
	assert args.K > 1, "Please select K > 1!"
	assert args.threads > 0, "Please select a valid number of threads!"
	assert args.seed >= 0, "Please select a valid seed!"
	assert args.iter > 0, "Please select a valid number of iterations!"
	assert args.tole >= 0.0, "Please select a valid tolerance!"
	assert args.check > 0, "Please select a valid value for convergence check!"
	start = time()

	# Control threads of external numerical libraries
	os.environ["MKL_NUM_THREADS"] = str(args.threads)
	os.environ["MKL_MAX_THREADS"] = str(args.threads)
	os.environ["OMP_NUM_THREADS"] = str(args.threads)
	os.environ["OMP_MAX_THREADS"] = str(args.threads)
	os.environ["NUMEXPR_NUM_THREADS"] = str(args.threads)
	os.environ["NUMEXPR_MAX_THREADS"] = str(args.threads)
	os.environ["OPENBLAS_NUM_THREADS"] = str(args.threads)
	os.environ["OPENBLAS_MAX_THREADS"] = str(args.threads)

	# Import numerical libraries and cython functions
	import numpy as np
	from hapla import functions
	from hapla import admix_cy

	# Prepare list of data files
	if args.filelist is not None:
		W = 0 # Counter for windows
		Z_list = [] # List of filenames
		with open(args.filelist) as f:
			for z_file in f:
				# Check input across files (chromosomes) and count windows
				z = z_file.strip("\n")
				Z_list.append(z)
				assert os.path.isfile(f"{z}.bca"), "bca file doesn't exist!"
				assert os.path.isfile(f"{z}.ids"), "ids file doesn't exist!"
				assert os.path.isfile(f"{z}.win"), "win file doesn't exist!"
				if W == 0: # First file
					z_ids = np.genfromtxt(f"{z}.ids", dtype=np.str_)
					k_vec = np.genfromtxt(f"{z}.win", dtype=np.uint8, usecols=[5])
					N = z_ids.shape[0]
					W = k_vec.shape[0]
					w_list = [W]
				else: # Loop files
					t_ids = np.genfromtxt(f"{z}.ids", dtype=np.str_)
					assert np.sum(z_ids != t_ids) == 0, "Samples don't match across files!"
					k_tmp = np.genfromtxt(f"{z}.win", dtype=np.uint8, usecols=[5])
					k_vec = np.append(k_vec, k_tmp)
					W += k_tmp.shape[0]
					w_list.append(k_tmp.shape[0])
		w_vec = np.array(w_list, dtype=np.uint32)
		del z_ids, t_ids, k_tmp, w_list
	else: # Single file (chromosome)
		Z_list = [args.clusters]
		assert os.path.isfile(f"{Z_list[0]}.bca"), "bca file doesn't exist!"
		assert os.path.isfile(f"{Z_list[0]}.ids"), "ids file doesn't exist!"
		assert os.path.isfile(f"{Z_list[0]}.win"), "win file doesn't exist!"
		k_vec = np.genfromtxt(f"{Z_list[0]}.win", dtype=np.uint8, usecols=[5])
		N = np.genfromtxt(f"{Z_list[0]}.ids", dtype=np.str_).shape[0]
		W = k_vec.shape[0]
		w_vec = np.array([W], dtype=np.uint32)
	c_vec = np.insert(np.cumsum(k_vec[:-1]*args.K, dtype=np.uint32), 0, 0)
	print(f"Parsing {len(Z_list)} file(s).")

	# Load haplotype cluster assignments from binary hapla format
	B = 0
	Z = np.zeros((W, 2*N), dtype=np.uint8)
	for z in np.arange(len(Z_list)):
		with open(f"{Z_list[z]}.bca", "rb") as f:
			# Check magic numbers
			m_vec = np.fromfile(f, dtype=np.uint8, count=3)
			assert np.allclose(m_vec, np.array([7, 9, 13], dtype=np.uint8)), "Magic number doesn't match file format!"
			
			# Add haplotype cluster assignments to container
			z_tmp = np.fromfile(f, dtype=np.uint8)
			z_tmp.shape = (w_vec[z], 2*N)
			Z[B:(B + w_vec[z]),:] = z_tmp
			B += w_vec[z]
		print(f"\rParsed file {z+1}/{len(Z_list)}", end="")
	del m_vec, z_tmp

	# Count haplotype cluster alleles
	M = np.sum(k_vec, dtype=np.uint32)

	# Print information
	print(f"\rLoaded haplotype cluster assignments:\n" + \
		f"- {N} samples\n" + \
		f"- {W} windows\n" + \
		f"- {M} clusters\n")
	print(f"Estimating ancestry proportions: K={args.K}, seed={args.seed}.")

	# Initialize parameters randomly
	rng = np.random.default_rng(args.seed)
	P = rng.random(size=(M*args.K)).clip(min=1e-5, max=1-(1e-5))
	Q = rng.random(size=(N, args.K)).clip(min=1e-5, max=1-(1e-5))
	Q /= np.sum(Q, axis=1, keepdims=True)
	admix_cy.createP(P, k_vec, c_vec, args.K)

	# Supervised setting
	if args.supervised is not None:
		print("Ancestry estimation in supervised mode!")
		y = np.genfromtxt(args.supervised, dtype=np.uint8).reshape(-1)
		assert y.shape[0] == N, f"Number of samples differ between files!"
		assert np.max(y) <= args.K, "Wrong number of ancestral sources!"
		assert np.min(y) >= 0, "Wrong format in population assignments!"
		print(f"{np.sum(y > 0)}/{N} samples with fixed ancestry.")
		admix_cy.superQ(Q, y)
	else:
		y = None
	
	# Set up containers for EM algorithm
	P1 = np.zeros_like(P)
	P2 = np.zeros_like(P)
	Q1 = np.zeros_like(Q)
	Q2 = np.zeros_like(Q)
	P_tmp = np.zeros_like(P)
	Q_tmp = np.zeros_like(Q)

	# Estimate initial log-likelihood
	L_pre = admix_cy.loglike(Z, P, Q, k_vec, c_vec)
	print(f"Initial log-like: {L_pre:.1f}\n")

	# Prime iterations
	for _ in np.arange(3):
		functions.steps(Z, P, Q, P_tmp, Q_tmp, k_vec, c_vec, y)

	# Accelerated EM algorithm
	ts = time()
	print(f"Accelerated EM algorithm.")
	for it in np.arange(args.iter):
		functions.quasi(Z, P, Q, P_tmp, Q_tmp, P1, P2, Q1, Q2, k_vec, c_vec, y)
		functions.steps(Z, P, Q, P_tmp, Q_tmp, k_vec, c_vec, y)

		# Log-likelihood convergence check
		if ((it+1) % args.check) == 0:
			L_cur = admix_cy.loglike(Z, P, Q, k_vec, c_vec)
			print(f"({it+1})\tLog-like: {L_cur:.1f}\t({time()-ts:.1f}s)", flush=True)
			if (abs(L_cur - L_pre) < args.tole):
				print("Converged!")
				print(f"Final log-likelihood: {L_cur:.1f}")
				break
			L_pre = L_cur
			ts = time()

	# Save output
	np.savetxt(f"{args.out}.K{args.K}.s{args.seed}.Q", Q, fmt="%.6f")
	print(f"Saved Q matrix as {args.out}.K{args.K}.s{args.seed}.Q")
	if not args.no_freq:
		if args.filelist is not None: # Save P file for each file (chromosome)
			w_cnt = c_cnt = 0
			for z in np.arange(len(Z_list)):
				p_num = np.sum(k_vec[w_cnt:(w_cnt + w_vec[z])], dtype=np.uint32)*args.K
				P[c_cnt:(c_cnt + p_num)].tofile(
					f"{args.out}.K{args.K}.s{args.seed}.file{z+1}.P.bin"
				)
				w_cnt += w_vec[z]
				c_cnt += p_num
			print(f"Saved P matrices (binary) as {args.out}.K{args.K}.s{args.seed}.file{{1..{len(Z_list)}}}.P.bin")
		else: # Single file (chromosome)
			P.tofile(f"{args.out}.K{args.K}.s{args.seed}.P.bin")
			print(f"Saved P matrix (binary) as {args.out}.K{args.K}.s{args.seed}.P.bin")

	# Print elapsed time for computation
	t_tot = time() - start
	t_min = int(t_tot//60)
	t_sec = int(t_tot - t_min*60)
	print(f"\nTotal elapsed time: {t_min}m{t_sec}s")



##### Main exception #####
assert __name__ != "__main__", "Please use the 'hapla admix' command!"



# cython: language_level=3, boundscheck=False, wraparound=False, initializedcheck=False, cdivision=True
cimport numpy as np
cimport openmp as omp
from cython.parallel import parallel, prange
from libc.math cimport log, log1p
from libc.stdint cimport uint8_t, uint32_t
from libc.stdlib cimport calloc, free

##### hapla - haplotype clustering #####
### Inline functions
# Calculate Hamming distance
cdef inline uint32_t _hammingDist(
		const uint8_t* X, const uint8_t* R, const size_t M
	) noexcept nogil:
	cdef:
		size_t j
		uint32_t dist = 0
	for j in range(M):
		if X[j] != R[j]:
			dist += 1
	return dist

# Calculate Hamming distance and change assignments
cdef inline uint32_t _hammingCheck(
		const uint8_t* z_vec, uint8_t* z_pre, const size_t U
	) noexcept nogil:
	cdef:
		size_t i
		uint32_t dist = 0
	for i in range(U):
		if z_vec[i] != z_pre[i]:
			z_pre[i] = z_vec[i]
			dist += 1
	return dist

# Add haplotype contribution to frequency vector
cdef inline void _addHaplo(
		const uint8_t* X, uint32_t* C, const uint32_t u, const size_t M
	) noexcept nogil:
	cdef size_t j
	for j in range(M):
		if X[j]:
			C[j] += u

# Update cluster information
cdef inline void _updateClust(
		const uint8_t* X, uint8_t* R, uint32_t* A, uint32_t* B, const uint32_t u, const size_t M
	) noexcept nogil:
	cdef:
		size_t j
	for j in range(M):
		R[j] = X[j]
		if X[j]:
			A[j] = u
			B[j] -= u
		else:
			A[j] = 0

# Estimate log-likelihood between cluster medians based on frequencies
cdef inline float _logLike(
		uint8_t* r, uint32_t* c, float n, size_t M
	) noexcept nogil:
	cdef:
		float s = 0.0
		float d, p
		size_t j
	for j in range(M):
		d = <float>r[j]
		p = <float>c[j]*n
		s += d*log(p) + (1.0 - d)*log1p(-p)
	return s

### Standard functions
# Create marginal medians
cpdef void marginalMedians(
		uint8_t[:,::1] R, uint32_t[:,::1] C, const uint32_t[::1] n_vec, const size_t K
	) noexcept nogil:
	cdef:
		size_t M = R.shape[1]
		size_t j, k
		uint32_t n
	for k in range(K):
		if n_vec[k] > 0:
			n = n_vec[k]//2
			for j in range(M):
				R[k,j] = C[k,j] > n
				C[k,j] = 0

# Compute distances, cluster assignment and prepare for next loop
cpdef void assignClust(
		uint8_t[:,::1] X, const uint8_t[:,::1] R, uint32_t[:,::1] C, uint8_t[::1] z_vec, uint32_t[::1] c_vec,
		const uint32_t[::1] n_vec, uint32_t[::1] n_tmp, const uint32_t[::1] u_vec, const size_t U, const size_t K
	) noexcept nogil:
	cdef:
		omp.omp_lock_t mutex
		size_t M = X.shape[1]
		size_t i, k, x, y, z
		uint8_t* xi
		uint32_t c, d, u
		uint32_t* C_thr
		uint32_t* n_thr
	omp.omp_init_lock(&mutex)
	with nogil, parallel():
		C_thr = <uint32_t*>calloc(K*M, sizeof(uint32_t))
		n_thr = <uint32_t*>calloc(K, sizeof(uint32_t))
		for i in prange(U):
			xi = &X[i,0]
			z = 0
			c = M + 1
			for k in range(K):
				if n_vec[k] > 0:
					d = _hammingDist(xi, &R[k,0], M)
					if d < c or (d == c and n_vec[k] > n_vec[z]):
						z = k
						c = d
			z_vec[i] = <uint8_t>z
			c_vec[i] = c

			# Add individual contributions to temporary arrays
			u = u_vec[i]
			n_thr[z] += u
			_addHaplo(xi, &C_thr[z*M], u, M)
		
		# omp critical
		omp.omp_set_lock(&mutex)
		for x in range(K):
			if n_vec[x] > 0:
				n_tmp[x] += n_thr[x]
				for y in range(M):
					C[x,y] += C_thr[x*M + y]
		omp.omp_unset_lock(&mutex)
		free(C_thr)
		free(n_thr)
	omp.omp_destroy_lock(&mutex)

# Copy and reset size of clusters
cpdef void updateN(
		uint32_t[::1] n_vec, uint32_t[::1] n_tmp, const size_t K
	) noexcept nogil:
	cdef size_t k
	for k in range(K):
		n_vec[k] = n_tmp[k]
		n_tmp[k] = 0

# Check and generate new cluster
cpdef uint32_t checkClust(
		const uint8_t[:,::1] X, uint8_t[:,::1] R, uint32_t[:,::1] C, uint8_t[::1] z_vec, const uint32_t[::1] c_vec,
		uint32_t[::1] n_vec, const uint32_t[::1] u_vec, const uint32_t c_lim, const size_t U, const size_t K
	) noexcept nogil:
	cdef:
		size_t M = X.shape[1]
		size_t L = R.shape[0]
		size_t c_arg = 0
		size_t i, z
		uint32_t c_max = c_vec[0]
		uint32_t u
	for i in range(1, U): # Find extreme point
		if c_vec[i] > c_max:
			c_arg = i
			c_max = c_vec[i]
	if c_max >= c_lim and K < L:
		u = u_vec[c_arg]
		z = <size_t>z_vec[c_arg]
		_updateClust(&X[c_arg,0], &R[K,0], &C[K,0], &C[z,0], u, M)
		n_vec[K] = u
		n_vec[z] -= u
		z_vec[c_arg] = K
		return 1
	else:
		return 0

# Generate new cluster from no check
cpdef void genClust(
		const uint8_t[:,::1] X, uint8_t[:,::1] R, uint32_t[:,::1] C, uint8_t[::1] z_vec, const uint32_t[::1] c_vec,
		uint32_t[::1] n_vec, const uint32_t[::1] u_vec, const size_t U, const size_t K
	) noexcept nogil:
	cdef:
		size_t M = X.shape[1]
		size_t c_arg = 0
		size_t i, z
		uint32_t c_max = c_vec[0]
		uint32_t u
	for i in range(1, U): # Find extreme point
		if c_vec[i] > c_max:
			c_arg = i
			c_max = c_vec[i]
	u = u_vec[c_arg]
	z = <size_t>z_vec[c_arg]
	_updateClust(&X[c_arg,0], &R[K,0], &C[K,0], &C[z,0], u, M)
	n_vec[K] = u
	n_vec[z] -= u
	z_vec[c_arg] = K
	
# Convergence check through hamming distance
cpdef uint32_t countDist(
		const uint8_t[::1] z_vec, uint8_t[::1] z_tmp, const size_t U
	) noexcept nogil:
	return _hammingCheck(&z_vec[0], &z_tmp[0], U)

# Set singleton clusters to zero
cpdef void setZero(
		uint32_t[::1] n_vec, const size_t lim, const size_t K
	) noexcept nogil:
	cdef:
		size_t k = K - 1
		uint32_t c_mac = 0
	while (c_mac < lim) and (k > 1):
		if n_vec[k] == 1:
			n_vec[k] = 0
			c_mac += 1
		k -= 1

# Find non-zero cluster with least assignments
cpdef uint32_t findZero(
		uint32_t[::1] n_vec, const uint32_t N, const uint32_t mac, const size_t K
	) noexcept nogil:
	cdef:
		size_t k
		uint32_t minI = K - 1
		uint32_t minN = N + 1
	for k in range(K):
		if n_vec[k] > 0 and n_vec[k] <= minN:
			minI = k
			minN = n_vec[k]
	if minN < mac:
		n_vec[minI] = 0
	return minN

# Fix index of medians
cpdef void medianFix(
		uint8_t[:,::1] R, uint8_t[::1] z_vec, uint32_t[::1] n_vec, const size_t K, const size_t U
	) noexcept nogil:
	cdef:
		size_t M = R.shape[1]
		size_t c = 0
		size_t i, j, k
	for k in range(K):
		if n_vec[k] > 0:
			if k != c:
				for j in range(M):
					R[c,j] = R[k,j]
				for i in range(U):
					if z_vec[i] == k:
						z_vec[i] = <uint8_t>c
				n_vec[c] = n_vec[k]
				n_vec[k] = 0
			c += 1

# Fix haplotype cluster assignments
cpdef void assignFix(
		uint8_t[:,::1] Z, uint8_t[::1] z_vec, uint32_t[::1] p_vec, uint32_t[::1] d_vec, const size_t w
	) noexcept nogil:
	cdef:
		size_t N = Z.shape[1]
		size_t u = 0
		size_t i
		uint8_t z = z_vec[0]
	for i in range(N):
		if d_vec[i] != 0:
			z = z_vec[u]
			u += 1
		Z[w,<size_t>p_vec[i]] = z

# Reset arrays for next iteration
cpdef void resetArrays(
		uint32_t[::1] c_vec, uint32_t[::1] n_vec, uint32_t[::1] p_vec, uint32_t[::1] d_vec, uint32_t[::1] u_vec
	) noexcept nogil:
	cdef:
		size_t N = c_vec.shape[0]
		size_t K = n_vec.shape[0]
		size_t i, k
	for k in range(K):
		n_vec[k] = 0
	for i in range(N):
		c_vec[i] = 0
		p_vec[i] = <uint32_t>i
		d_vec[i] = 0
		u_vec[i] = 0

# Estimate log-likelihoods between cluster medians
cpdef void estimateLoglike(
		uint8_t[:,::1] R, uint32_t[:,::1] C, float[:,::1] L, uint32_t[::1] n_vec, size_t K
	) noexcept nogil:
	cdef:
		size_t M = R.shape[1]
		size_t k1, k2
	for k1 in range(K):
		for k2 in range(K):
			L[k1,k2] = _logLike(&R[k1,0], &C[k2,0], 1.0/<float>n_vec[k2], M)



# cython: language_level=3, boundscheck=False, wraparound=False, initializedcheck=False, cdivision=True
cimport numpy as np
from cython.parallel import parallel, prange
from libc.stdint cimport uint8_t, int16_t, uint32_t
from libc.stdlib cimport malloc, free

##### Cython functions for reading genotype files #####
# Read variant from VCF/BCF into 8-bit integer format
cpdef void readVar(
		uint8_t[::1] G, const int16_t[:,::1] V, const size_t N
	) noexcept nogil:
	cdef size_t i
	for i in range(N):
		G[2*i] = <uint8_t>V[i,0] # Allele 1
		G[2*i+1] = <uint8_t>V[i,1] # Allele 2

# Read variant from VCF/BCF into 8-bit integer format with missing option
cpdef void predVar(
		uint8_t[::1] G, const int16_t[:,::1] V, const size_t N
	) noexcept nogil:
	cdef size_t i
	for i in range(N):
		G[2*i] = 9 if V[i,0] == -1 else <uint8_t>V[i,0] # Allele 1
		G[2*i+1] = 9 if V[i,1] == -1 else <uint8_t>V[i,1] # Allele 2

# Initialize cluster mean and suffix arrays
cpdef void convertHap(
		const uint8_t[:,::1] G, uint32_t[:,::1] C, uint32_t[::1] p_vec, uint32_t[::1] d_vec, uint32_t[::1] a_tmp, 
		uint32_t[::1] b_tmp, uint32_t[::1] d_tmp, uint32_t[::1] e_tmp, const size_t S
	) noexcept nogil:
	cdef:
		size_t M = C.shape[1]
		size_t N = G.shape[1]
		size_t i, j, k, s, u, v
		uint32_t f, l, p, q
	for j in range(M):
		s = S + j
		u = v = 0
		p = q = <uint32_t>j + 1
		C[0,j] = 0
		for i in range(N):
			# Add to cluster mean
			C[0,j] += G[s,i]

			# Suffix array updates
			f = p_vec[i]
			l = d_vec[i]
			if l > p:
				p = l
			if l > q:
				q = l
			if G[s,f] == 0:
				a_tmp[u] = f
				d_tmp[u] = p
				u += 1
				p = 0
			else:
				b_tmp[v] = f
				e_tmp[v] = q
				v += 1
				q = 0
		for k in range(u):
			p_vec[k] = a_tmp[k]
			d_vec[k] = d_tmp[k]
		for k in range(v):
			p_vec[u+k] = b_tmp[k]
			d_vec[u+k] = e_tmp[k]

# Extract unique haplotypes from suffix arrays
cpdef uint32_t uniqueHap(
		const uint8_t[:,::1] G, uint8_t[:,::1] X, const uint32_t[::1] p_vec, const uint32_t[::1] d_vec, 
		uint32_t[::1] u_vec, const size_t S
	) noexcept nogil:
	cdef:
		size_t N = X.shape[0]
		size_t M = X.shape[1]
		size_t u = 0
		size_t h, i, j
	for i in range(N):
		if d_vec[i] != 0:
			h = <size_t>p_vec[i]
			for j in range(M):
				X[u,j] = G[S+j,h]
			u += 1
		u_vec[u - 1] += 1
	return u
			
# Convert transposed window for predicting target clusters
cpdef void predictHap(
		const uint8_t[:,::1] G, uint8_t[:,::1] X, const size_t S
	) noexcept nogil:
	cdef:
		size_t N = X.shape[0]
		size_t M = X.shape[1]
		size_t i, j
	for i in range(N):
		for j in range(M):
			X[i,j] = G[S+j,i]

# Convert haplotype cluster alleles to 2-bit PLINK format
cpdef void convertPlink(
		const uint8_t[:,::1] Z, uint8_t[:,::1] Z_bin, uint32_t[:,::1] P_mat, const uint8_t[::1] k_vec, 
		const uint32_t[::1] c_vec, const uint32_t[::1] b_vec
	) noexcept nogil:
	cdef:
		size_t W = Z.shape[0]
		size_t N = Z.shape[1]//2
		size_t B = Z_bin.shape[1]
		size_t b, c, i, j, l, n, s, w, bit
		uint8_t* z_vec
	with nogil, parallel():
		z_vec = <uint8_t*>malloc(sizeof(uint8_t)*N)
		for w in prange(W):
			s = <size_t>c_vec[w]
			for c in range(k_vec[w]):
				# Create haplotype cluster alleles
				l = s + c
				for i in range(0, 2*N, 2):
					n = i//2
					z_vec[n] = 0
					if Z[w,i] == c:
						z_vec[n] += 1
					if Z[w,i+1] == c:
						z_vec[n] += 1

				# Save in 2-bit form with bit-wise operations
				j = 0
				for b in range(B):
					for bit in range(0, 8, 2):
						if z_vec[j] == 0:
							Z_bin[l,b] |= (1<<bit)
							Z_bin[l,b] |= (1<<(bit+1))
						if z_vec[j] == 1:
							Z_bin[l,b] |= (1<<(bit+1))

						# Increase counter and check for break
						j = j + 1
						if j == N:
							break
				
				# Save window and cluster information
				P_mat[l,0] = <uint32_t>(w + 1)
				P_mat[l,1] = <uint32_t>(c + 1)
				P_mat[l,2] = b_vec[w]
		free(z_vec)

# Convert 2-bit into standardized genotype array for phenotypes
cpdef void phenoPlink(
		const uint8_t[:,::1] G_mat, double[:,::1] G, const uint32_t[::1] c
	) noexcept nogil:
	cdef:
		size_t M = G.shape[0]
		size_t N = G.shape[1]
		size_t B = G_mat.shape[1]
		size_t b, i, j, bytepart
		uint8_t[4] recode = [2, 9, 1, 0]
		uint8_t mask = 3
		uint8_t byte
	for j in range(M):
		i = 0
		for b in range(B):
			byte = G_mat[<size_t>c[j],b]
			for bytepart in range(4):
				G[j,i] = <double>recode[byte & mask]
				byte = byte >> 2
				i += 1
				if i == N:
					break



"""
hapla.
Infer local ancestry tracts.
"""

__author__ = "Jonas Meisner"

# Libraries
import os
from time import time

##### hapla fatash #####
def main(args):
	print("-----------------------------------")
	print("hapla by Jonas Meisner (v0.23.0)")
	print(f"hapla fatash using {args.threads} thread(s)")
	print("-----------------------------------\n")

	# Check input
	assert (args.filelist is not None) or (args.clusters is not None), "No input data (--filelist or --clusters)!"
	assert (args.pfilelist is not None) or (args.pfile is not None), "No P files provided (--pfilelist or --pfile)!"
	if args.filelist is not None:
		assert args.pfilelist is not None, "Input formats don't match!"
	if args.clusters is not None:
		assert args.pfile is not None, "Input formats don't match!"
	assert args.qfile is not None, "No Q-matrix provided (--q-matrix)!"
	assert args.threads > 0, "Please select a valid number of threads!"
	assert args.alpha > 0.0, "Please select a valid alpha value!"
	start = time()

	# Control threads of external numerical libraries
	os.environ["MKL_NUM_THREADS"] = str(args.threads)
	os.environ["MKL_MAX_THREADS"] = str(args.threads)
	os.environ["OMP_NUM_THREADS"] = str(args.threads)
	os.environ["OMP_MAX_THREADS"] = str(args.threads)
	os.environ["NUMEXPR_NUM_THREADS"] = str(args.threads)
	os.environ["NUMEXPR_MAX_THREADS"] = str(args.threads)
	os.environ["OPENBLAS_NUM_THREADS"] = str(args.threads)
	os.environ["OPENBLAS_MAX_THREADS"] = str(args.threads)

	# Import numerical libraries and cython functions
	import numpy as np
	from hapla import fatash_cy

	# Prepare list of data files
	if args.filelist is not None:
		Z_list = [] # List of filenames
		with open(args.filelist) as f:
			for z_file in f:
				# Check input across files and count windows
				z = z_file.strip("\n")
				Z_list.append(z)
				assert os.path.isfile(f"{z}.bca"), "bca file doesn't exist!"
				assert os.path.isfile(f"{z}.ids"), "ids file doesn't exist!"
				assert os.path.isfile(f"{z}.win"), "win file doesn't exist!"
				if len(Z_list) == 1: # First file
					z_ids = np.genfromtxt(f"{z}.ids", dtype=np.str_)
					k_vec = np.genfromtxt(f"{z}.win", dtype=np.uint8, usecols=[5])
					N = 2*z_ids.shape[0]
					w_list = [k_vec.shape[0]]
				else: # Loop files
					t_ids = np.genfromtxt(f"{z}.ids", dtype=np.str_)
					assert np.sum(z_ids != t_ids) == 0, "Samples do not match across files!"
					k_tmp = np.genfromtxt(f"{z}.win", dtype=np.uint8, usecols=[5])
					k_vec = np.append(k_vec, k_tmp)
					w_list.append(k_tmp.shape[0])
		w_vec = np.array(w_list, dtype=int)
		del z_ids, t_ids, k_tmp, w_list
	else: # Single file (chromosome)
		Z_list = [args.clusters]
		assert os.path.isfile(f"{Z_list[0]}.bca"), "bca file doesn't exist!"
		assert os.path.isfile(f"{Z_list[0]}.ids"), "ids file doesn't exist!"
		assert os.path.isfile(f"{Z_list[0]}.win"), "win file doesn't exist!"
		k_vec = np.genfromtxt(f"{Z_list[0]}.win", dtype=np.uint8, usecols=[5])
		w_vec = np.array([k_vec.shape[0]], dtype=int)
		N = 2*np.genfromtxt(f"{Z_list[0]}.ids", dtype=np.str_).shape[0]
	print(f"Parsing {len(Z_list)} file(s).")

	# Load Q matrix
	Q = np.genfromtxt(args.qfile, dtype=float)
	if Q.shape[0] == N//2:
		Q = np.repeat(Q, 2, axis=0)
	else:
		assert Q.shape[0] == N, "Number of samples doesn't match!"
	Q_log = np.log(Q)
	K = Q.shape[1]

	# Prepare list of P files
	if args.pfilelist is not None:
		w_cnt = 0
		P_list = []
		with open(args.pfilelist) as f:
			for p_file in f:
				P_list.append(p_file.strip("\n"))
		assert len(P_list) == len(Z_list), "Number of files doesn't match!"

	# General containers
	T = np.zeros((K, K)) # Transitions
	v = np.zeros(K) # Help vector

	# Loop over chromosomes
	print(f"Inferring local ancestry tracts with {K} ancestral sources.\n")
	for z in np.arange(len(Z_list)): # Loop through files
		print(f"Processing file {z+1}/{len(Z_list)}")
		s_tmp = time()

		# Load P matrix file
		if len(Z_list) > 1:
			P_tmp = np.fromfile(P_list[z], dtype=float)
			k_tmp = k_vec[w_cnt:(w_cnt + w_vec[z])]
			w_cnt += w_vec[z]
		else:
			P_tmp = np.fromfile(args.pfile, dtype=float)
			k_tmp = k_vec
		c_tmp = np.insert(np.cumsum(k_tmp[:-1]*K, dtype=np.uint32), 0, 0)
		p_num = np.sum(k_tmp, dtype=int)*K
		assert P_tmp.shape[0] == p_num, "Number of clusters doesn't match!"
		assert (P_tmp.shape[0] % K) == 0, "Parameters don't match!"

		# Load haplotype cluster assignment file
		with open(f"{Z_list[z]}.bca", "rb") as f:
			# Check magic numbers
			m_vec = np.fromfile(f, dtype=np.uint8, count=3)
			assert np.allclose(m_vec, np.array([7, 9, 13], dtype=np.uint8)), "Magic number doesn't match file format!"
			
			# Add haplotype cluster assignments to container
			Z_tmp = np.fromfile(f, dtype=np.uint8)
			Z_tmp.shape = (w_vec[z], N)
		Z_tmp = np.ascontiguousarray(Z_tmp.T) # Transpose for easier computations
		assert np.max(k_tmp) == (np.max(Z_tmp) + 1), "Number of clusters doesn't match!"

		# Containers
		E = np.zeros((N, w_vec[z], K)) # Emission probabilities
		A = np.zeros((w_vec[z], K)) # Forward matrix
		if args.viterbi:
			I = np.zeros((w_vec[z], K), dtype=np.uint8) # Index matrix
			V = np.zeros((N, w_vec[z]), dtype=np.uint8) # Viterbi path
		else:
			B = np.zeros((w_vec[z], K)) # Backward matrix
			L = np.zeros((N, w_vec[z], K)) # Posterior probabilities

		# Compute emission probabilities
		fatash_cy.calcEmissions(E, Z_tmp, P_tmp, k_tmp, c_tmp)
		del Z_tmp, P_tmp, k_tmp, c_tmp

		# HMM for each haplotype
		for i in np.arange(N):
			print(f"\rHaplotype {i+1}/{N}", end="")
			fatash_cy.calcTransition(T, Q[i,:], args.alpha)
			if args.viterbi: # Compute Viterbi and decoding
				fatash_cy.viterbi(E[i,:,:], Q_log[i,:], T, A, I, V[i,:])
			else: # Compute posterior probabilities
				fatash_cy.calcFwdBwd(E[i,:,:], L[i,:,:], Q_log[i,:], T, A, B, v)
		print(".")

		# Save matrices
		if len(Z_list) == 1:
			if args.viterbi:
				np.savetxt(f"{args.out}.path", V, fmt="%i")
				print(f"Saved Viterbi decoding path as {args.out}.path")
			else:
				np.savetxt(f"{args.out}.path", L.argmax(axis=2), fmt="%i")
				print(f"Saved posterior decoding path as {args.out}.path")
				if args.save_posterior:
					np.save(f"{args.out}.post", L)
					print(f"Saved posteriors as {args.out}.post.npy")
			print("")
		else:
			if args.viterbi:
				np.savetxt(f"{args.out}.file{z+1}.path", V, fmt="%i")
				print(f"Saved Viterbi decoding as {args.out}.file{z+1}.path")
			else:
				np.savetxt(f"{args.out}.file{z+1}.path", L.argmax(axis=2), fmt="%i")
				print(f"Saved posterior decoding as {args.out}.file{z+1}.path")
				if args.save_posterior:
					L.tofile(f"{args.out}.file{z+1}.post.bin")
					print(f"Saved posteriors (binary) as {args.out}.file{z+1}.post.bin")

			# Print elapsed time of chromosome 
			t_tmp = time()-s_tmp
			t_min = int(t_tmp//60)
			t_sec = int(t_tmp - t_min*60)
			print(f"Elapsed time: {t_min}m{t_sec}s\n")
		del E, A
		if args.viterbi:
			del I, V
		else:
			del B, L

	# Print elapsed time for computation
	t_tot = time() - start
	t_min = int(t_tot//60)
	t_sec = int(t_tot - t_min*60)
	print(f"Total elapsed time: {t_min}m{t_sec}s")



##### Main exception #####
assert __name__ != "__main__", "Please use the 'hapla fatash' command!"



"""
hapla.
Population structure inference using haplotype cluster alleles.
"""

__author__ = "Jonas Meisner"

# Libraries
import os
from time import time

##### hapla struct #####
def main(args):
	print("-----------------------------------")
	print("hapla by Jonas Meisner (v0.23.0)")
	print(f"hapla struct using {args.threads} thread(s)")
	print("-----------------------------------\n")

	# Check input
	assert (args.filelist is not None) or (args.clusters is not None), "No input data (--filelist or --clusters)!"
	assert (args.grm or (args.pca is not None)), "No analysis selected (--grm, --pca)!"
	if args.pca is not None:
		assert args.pca > 0, "Please select a valid number of eigenvectors!"
	assert args.threads > 0, "Please select a valid number of threads!"
	assert args.batch > 0, "Please select a valid batch size!"
	assert args.power > 0, "Please select a valid number of power iterations!"
	assert args.seed >= 0, "Please select a valid seed!"
	start = time()

	# Control threads of external numerical libraries
	os.environ["MKL_NUM_THREADS"] = str(args.threads)
	os.environ["MKL_MAX_THREADS"] = str(args.threads)
	os.environ["OMP_NUM_THREADS"] = str(args.threads)
	os.environ["OMP_MAX_THREADS"] = str(args.threads)
	os.environ["NUMEXPR_NUM_THREADS"] = str(args.threads)
	os.environ["NUMEXPR_MAX_THREADS"] = str(args.threads)
	os.environ["OPENBLAS_NUM_THREADS"] = str(args.threads)
	os.environ["OPENBLAS_MAX_THREADS"] = str(args.threads)

	# Import numerical libraries and cython functions
	import numpy as np
	from math import ceil
	from hapla import functions
	from hapla import shared_cy

	# Prepare list of data files
	if args.filelist is not None:
		W = 0 # Counter for windows
		Z_list = [] # List of filenames
		with open(args.filelist) as f:
			for z_file in f:
				# Check input across files and count windows
				z = z_file.strip("\n")
				Z_list.append(z)
				assert os.path.isfile(f"{z}.bca"), "bca file doesn't exist!"
				assert os.path.isfile(f"{z}.ids"), "ids file doesn't exist!"
				assert os.path.isfile(f"{z}.win"), "win file doesn't exist!"
				if W == 0: # First file
					z_ids = np.genfromtxt(f"{z}.ids", dtype=np.str_)
					k_vec = np.genfromtxt(f"{z}.win", dtype=np.uint8, usecols=[5])
					N = z_ids.shape[0]
					W = k_vec.shape[0]
					w_list = [W]
				else: # Loop files
					t_ids = np.genfromtxt(f"{z}.ids", dtype=np.str_)
					assert np.sum(z_ids != t_ids) == 0, "Samples don't match across files!"
					k_tmp = np.genfromtxt(f"{z}.win", dtype=np.uint8, usecols=[5])
					k_vec = np.append(k_vec, k_tmp)
					W += k_tmp.shape[0]
					w_list.append(k_tmp.shape[0])
		w_vec = np.array(w_list, dtype=np.uint32)
		del t_ids, k_tmp, w_list
	else: # Single file (chromosome)
		Z_list = [args.clusters]
		assert os.path.isfile(f"{Z_list[0]}.bca"), "bca file doesn't exist!"
		assert os.path.isfile(f"{Z_list[0]}.ids"), "ids file doesn't exist!"
		assert os.path.isfile(f"{Z_list[0]}.win"), "win file doesn't exist!"
		z_ids = np.genfromtxt(f"{Z_list[0]}.ids", dtype=np.str_)
		k_vec = np.genfromtxt(f"{Z_list[0]}.win", dtype=np.uint8, usecols=[5])
		N = z_ids.shape[0]
		W = k_vec.shape[0]
		w_vec = np.array([W], dtype=np.uint32)
	print(f"Parsing {len(Z_list)} file(s).")
	
	# Estimate genome-wide relationship matrix
	if args.grm:
		print("Estimating genome-wide relationship matrix (GRM).")
		s_pre = ""
		K = 0
		D = 0.0
		G = np.zeros((N, N), dtype=np.float32)
		for z in np.arange(len(Z_list)): # Loop through files
			# Print information
			s_beg = f"Processing file {z+1}/{len(Z_list)}"
			print(f"\r{s_beg: <{len(s_pre)}}", end="")

			# Load haplotype cluster assignment file
			with open(f"{Z_list[z]}.bca", "rb") as f:
				# Check magic numbers
				m_vec = np.fromfile(f, dtype=np.uint8, count=3)
				assert np.allclose(m_vec, np.array([7, 9, 13], dtype=np.uint8)), \
					"Magic number doesn't match file format!"
				
				# Add haplotype cluster assignments to container
				Z_tmp = np.fromfile(f, dtype=np.uint8)
				Z_tmp.shape = (w_vec[z], 2*N)
			del m_vec

			# File setup
			k_tmp = k_vec[K:(K + w_vec[z])]
			c_tmp = np.insert(np.cumsum(k_tmp[:-1], dtype=np.uint32), 0, 0)
			M = np.sum(k_tmp, dtype=np.uint32)

			# Aggregate alleles and estimate cluster frequencies
			Z_agg = np.zeros((M, N), dtype=np.uint8)
			p_tmp = np.zeros(M)
			shared_cy.haplotypeAggregate(Z_tmp, Z_agg, p_tmp, k_tmp, c_tmp)
			del Z_tmp, k_tmp, c_tmp

			# Setup GRM part settings
			B = ceil(M/args.batch) # Number of batches
			Z_bat = np.zeros((args.batch, N), dtype=np.float32)

			# Estimate GRM part in batches
			for b in np.arange(B):
				s_bat = f"{s_beg}. Batch {b+1}/{B}"
				print(f"\r{s_bat}", end="") # Print information
				M_b = b*args.batch
				if b == (B-1): # Last batch
					Z_bat = np.zeros((M - M_b, N), dtype=np.float32)
				shared_cy.centerZ(Z_agg, Z_bat, p_tmp, M_b)

				# Aggregate across batches
				G += np.dot(Z_bat.T, Z_bat)
			K += w_vec[z]
			D += np.sum(p_tmp*(1.0 - p_tmp), dtype=float)
			s_pre = s_bat
			del p_tmp, Z_agg, Z_bat
		G *= (1.0/(2.0*D))
		print(".\n")
		
		# Gower and data centering of GRM
		if not args.no_centering:
			print("Centering GRM.")
			u = np.mean(G, axis=1)
			G -= u.reshape(1, N)
			u = np.mean(G, axis=1)
			G -= u.reshape(N, 1)

			# Gower centering
			G *= float(N-1)/np.trace(G)
			del u

		# Save matrix
		G = G[np.tril_indices(N)]
		G.tofile(f"{args.out}.grm.bin")
		np.full(N*(N+1)//2, D, dtype=np.float32).tofile(f"{args.out}.grm.N.bin")
		z_ids = z_ids.reshape(-1, 1)
		if args.duplicate_fid:
			fam = z_ids.repeat(2, axis=1)
		else:
			fam = np.hstack((np.zeros((N, 1), dtype=np.uint8), z_ids))
		np.savetxt(f"{args.out}.grm.id", fam, delimiter="\t", fmt="%s")
		print("Saved genome-wide relationship matrix in GCTA format:\n" + \
			f"- {args.out}.grm.bin\n" + \
			f"- {args.out}.grm.N.bin\n" + \
			f"- {args.out}.grm.id\n")
		del G, fam
	
	# Infer population structure using PCA
	if args.pca is not None:
		# Load haplotype cluster assignments from binary hapla format
		B = 0
		Z = np.zeros((W, 2*N), dtype=np.uint8)
		for z in np.arange(len(Z_list)):
			with open(f"{Z_list[z]}.bca", "rb") as f:
				# Check magic numbers
				m_vec = np.fromfile(f, dtype=np.uint8, count=3)
				assert np.allclose(m_vec, np.array([7, 9, 13], dtype=np.uint8)), \
					"Magic number doesn't match file format!"
				
				# Add haplotype cluster assignments to container
				z_tmp = np.fromfile(f, dtype=np.uint8)
				z_tmp.shape = (w_vec[z], 2*N)
				Z[B:(B + w_vec[z]),:] = z_tmp
				B += w_vec[z]
			print(f"\rParsed file {z+1}/{len(Z_list)}", end="")
		del m_vec, z_tmp

		# Count haplotype cluster alleles
		M = np.sum(k_vec, dtype=np.uint32)

		# Print information
		print(f"\rLoaded haplotype cluster assignments:\n" + \
			f"- {N} samples\n" + \
			f"- {W} windows\n" + \
			f"- {M} clusters\n")
		
		# Populate full matrix and estimate cluster frequencies
		Z_agg = np.zeros((M, N), dtype=np.uint8)
		p_vec = np.zeros(M)
		c_vec = np.insert(np.cumsum(k_vec[:-1], dtype=np.uint32), 0, 0)
		shared_cy.haplotypeAggregate(Z, Z_agg, p_vec, k_vec, c_vec)
		del Z, k_vec, c_vec
		a_vec = 1.0/np.sqrt(2.0*p_vec*(1.0 - p_vec))

		# Randomized SVD
		print(f"Computing randomized SVD, extracting {args.pca} eigenvectors.")
		rng = np.random.default_rng(args.seed)
		U, S, V = functions.randomizedSVD(Z_agg, p_vec, a_vec, args.pca, args.batch, args.power, rng)
		print(".\n")

		# Save matrices
		if args.raw: # Only eigenvectors
			np.savetxt(f"{args.out}.eigenvecs", V, fmt="%.7f")
		else: # Include FID and IID fields
			z_ids = z_ids.reshape(-1, 1)
			if args.duplicate_fid:
				fam = z_ids.repeat(2, axis=1)
			else:
				fam = np.hstack((np.zeros((N, 1), dtype=np.uint8), z_ids))
			V = np.hstack((fam, np.round(V, 7)))
			h = ["#FID", "IID"] + [f"PC{k}" for k in range(1, args.pca+1)]
			np.savetxt(f"{args.out}.eigenvecs", V, fmt="%s", delimiter="\t", comments="", header="\t".join(h))
			print(f"Saved eigenvectors as {args.out}.eigenvecs")
		np.savetxt(f"{args.out}.eigenvals", (S*S)/float(M), fmt="%.7f")
		print(f"Saved eigenvalues as {args.out}.eigenvals")
		if args.loadings:
			np.savetxt(f"{args.out}.loadings", U, fmt="%.7f")
			print(f"Saved loadings as {args.out}.loadings")
		print("")
		del z_ids, Z_agg, p_vec, a_vec, U, S, V, h, fam

	# Print elapsed time for computation
	t_tot = time() - start
	t_min = int(t_tot//60)
	t_sec = int(t_tot - t_min*60)
	print(f"Total elapsed time: {t_min}m{t_sec}s")



##### Main exception #####
assert __name__ != "__main__", "Please use the 'hapla struct' command!"



"""
hapla.
Haplotype clustering using K-Medians with delayed cluster creation.
"""

__author__ = "Jonas Meisner"

# Libraries
import os
from time import time

##### hapla cluster #####
def main(args):
	print("-----------------------------------")
	print("hapla by Jonas Meisner (v0.23.0)")
	print(f"hapla cluster using {args.threads} thread(s)")
	print("-----------------------------------\n")

	# Check input
	assert args.vcf is not None, "No phased genotype file (--bcf or --vcf)!"
	assert os.path.isfile(f"{args.vcf}"), "VCF/BCF file doesn't exist!"
	assert os.path.isfile(f"{args.vcf}.csi") or os.path.isfile(f"{args.vcf}.tbi"), "VCF/BCF index doesn't exist!"
	assert args.threads > 0, "Please select a valid number of threads!"
	assert args.min_freq > 0.0, "Invalid haplotype cluster frequency!"
	if args.min_mac is not None:
		assert args.min_mac > 1, "Please select a valid MAC threshold!"
	assert args.max_iterations > 0, "Please select a valid number of iterations!"
	assert (args.lmbda > 0.0) and (args.lmbda < 1.0), "Please select a valid lambda value!"
	assert (args.max_clusters > 1) and (args.max_clusters <= 256), "Max allowed clusters exceeded!"
	if args.size is not None:
		assert args.size > 0, "Invalid window size!"
		if args.step is not None:
			if args.size == 1:
				args.step = None
			else:
				assert (args.step <= args.size) and (args.step > 0), "Invalid step size for sliding window chosen!"
	else:
		assert args.windows is not None, "No window option (--size or --windows)!"
	start = time()

	# Control threads of external numerical libraries
	os.environ["MKL_NUM_THREADS"] = str(args.threads)
	os.environ["MKL_MAX_THREADS"] = str(args.threads)
	os.environ["OMP_NUM_THREADS"] = str(args.threads)
	os.environ["OMP_MAX_THREADS"] = str(args.threads)
	os.environ["NUMEXPR_NUM_THREADS"] = str(args.threads)
	os.environ["NUMEXPR_MAX_THREADS"] = str(args.threads)
	os.environ["OPENBLAS_NUM_THREADS"] = str(args.threads)
	os.environ["OPENBLAS_MAX_THREADS"] = str(args.threads)

	# Import numerical libraries and cython functions
	import numpy as np
	from cyvcf2 import VCF
	from math import ceil
	from hapla import reader_cy
	from hapla import memory_cy
	from hapla import cluster_cy

	# Initiate VCF and extract parameters
	print("\rLoading VCF/BCF file...", end="")
	v_file = VCF(args.vcf, threads=min(args.threads, 4))
	s_list = np.array(v_file.samples).reshape(-1, 1)
	N = 2*s_list.shape[0]
	M = v_file.num_records
	B = ceil(N/8)

	# Set haplotype cluster size threshold
	if args.min_mac is not None:
		N_mac = args.min_mac
	else:
		N_mac = ceil(N*args.min_freq)

	# Allocate arrays
	if args.memory:
		G = np.zeros((M, B), dtype=np.uint8)
	else:
		G = np.zeros((M, N), dtype=np.uint8)
	v_vec = np.zeros(M, dtype=np.uint32)

	# Read variants into matrix
	for j, variant in enumerate(v_file):
		V = variant.genotype.array()
		if args.memory:
			memory_cy.readBit(G[j], V, N//2)
		else:
			reader_cy.readVar(G[j], V, N//2)
		v_vec[j] = variant.POS
	chrom = variant.CHROM # Extract chromosome information
	del V, v_file
	t_par = time() - start
	print(f"\rLoaded phased genotype data: {N} haplotypes and {M} SNPs.")

	# Set up windows
	if args.size is not None:
		if args.step is not None:
			W = ceil((M - args.size)/args.step)
			w_vec = [w*args.step for w in range(W)]
			print(f"Clustering {W} overlapping windows of {args.size} SNPs " + \
		 			f"(step-size {args.step}).")
		else:
			W = M//args.size
			w_vec = [w*args.size for w in range(W)]
			print(f"Clustering {W} non-overlapping windows of {args.size} SNPs.")
		w_vec.append(M)
		w_vec = np.array(w_vec, dtype=np.uint32)
	else:
		w_vec = np.genfromtxt(args.windows, dtype=np.uint32)
		assert w_vec[-1] <= M, "Genotype and window files don't match!"
		if w_vec[-1] != M:
			w_vec = np.insert(w_vec, W, M)
		W = w_vec.shape[0]
		print(f"Clustering {W} windows with provided SNP lengths.")

	# Containers
	Z = np.zeros((W, N), dtype=np.uint8) # Chromosome-based cluster assignments
	z_vec = np.zeros(N, dtype=np.uint8) # Window-based cluster assignments 
	k_vec = np.zeros(W, dtype=np.uint8) # Number of clusters in windows
	c_vec = np.zeros(N, dtype=np.uint32) # Cost vector
	u_vec = np.zeros(N, dtype=np.uint32) # Count of unique haplotypes
	d_vec = np.zeros(N, dtype=np.uint32) # Divergence vector (suffix array)
	p_vec = np.arange(N, dtype=np.uint32) # Prefix vector (suffix array)
	n_vec = np.zeros(args.max_clusters, dtype=np.uint32) # Size vector
	z_tmp = np.zeros_like(z_vec) # Help vector (clustering)
	a_tmp = np.zeros_like(p_vec) # Help vector (suffix array)
	b_tmp = np.zeros_like(p_vec) # Help vector (suffix array)
	d_tmp = np.zeros_like(d_vec) # Help vector (suffix array)
	e_tmp = np.zeros_like(d_vec) # Help vector (suffix array)
	n_tmp = np.zeros_like(n_vec) # Help vector (size)
	if args.size is not None: # Window length-based
		if args.memory:
			H = np.zeros((args.size, N), dtype=np.uint8) # Haplotypes transposed
		X = np.zeros((N, args.size), dtype=np.uint8) # Haplotypes
		R = np.zeros((args.max_clusters, args.size), dtype=np.uint8) # Medians
		C = np.zeros((args.max_clusters, args.size), dtype=np.uint32) # Means
		c_lim = np.uint32(args.lmbda*float(X.shape[1])) # SNP-based threshold

	# Optional containers
	if args.medians:
		N_arr = np.array([], dtype=np.uint32) # Haplotype cluster sizes
		L = np.zeros((args.max_clusters, args.max_clusters), dtype=np.float32) # Log-likelihoods
		with open(f"{args.out}.bcm", "wb") as f: # Medians file
			np.array([7, 9, 13], dtype=np.uint8).tofile(f)
		with open(f"{args.out}.blk", "wb") as f: # Log-likelihoods file
			np.array([7, 9, 13], dtype=np.uint8).tofile(f)
		np.savetxt(f"{args.out}.wix", w_vec[:-1], fmt="%i") # Window lengths

	# Clustering using PDC-DP-Medians
	for w in np.arange(W):
		S = w_vec[w]
		print(f"\rWindow {w+1}/{W}", end="")

		# Prepare containers if window indices provided
		if args.size is None:
			if args.memory:
				H = np.zeros((w_vec[w+1]-S, N), dtype=np.uint8)
			X = np.zeros((N, w_vec[w+1]-S), dtype=np.uint8)
			R = np.zeros((args.max_clusters, X.shape[1]), dtype=np.uint8)
			C = np.zeros((args.max_clusters, X.shape[1]), dtype=np.uint32)
			c_lim = np.uint32(args.lmbda*float(X.shape[1]))

		# Prepare last window
		if w == (W-1):
			if args.memory:
				H = np.zeros((M-S, N), dtype=np.uint8)
			X = np.zeros((N, M-S), dtype=np.uint8)
			R = np.zeros((args.max_clusters, X.shape[1]), dtype=np.uint8)
			C = np.zeros((args.max_clusters, X.shape[1]), dtype=np.uint32)
			c_lim = np.uint32(args.lmbda*float(X.shape[1]))

		# Load haplotype window
		if args.memory:
			memory_cy.convertBit(G, H, C, p_vec, d_vec, a_tmp, b_tmp, d_tmp, e_tmp, S)
			U = memory_cy.uniqueBit(H, X, p_vec, d_vec, u_vec)
		else:
			reader_cy.convertHap(G, C, p_vec, d_vec, a_tmp, b_tmp, d_tmp, e_tmp, S)
			U = reader_cy.uniqueHap(G, X, p_vec, d_vec, u_vec, S)

		# Compute mean and initialize first median
		K = np.uint32(1)
		n_vec[0] = N
		cluster_cy.marginalMedians(R, C, n_vec, K)

		# Perform PDC-DP-Medians
		for it in np.arange(args.max_iterations):
			cluster_cy.assignClust(X, R, C, z_vec, c_vec, n_vec, n_tmp, u_vec, U, K)
			cluster_cy.updateN(n_vec, n_tmp, K)
			K += cluster_cy.checkClust(X, R, C, z_vec, c_vec, n_vec, u_vec, c_lim, U, K)

			# Check for convergence
			if it > 0:
				if cluster_cy.countDist(z_vec, z_tmp, U) == 0:
					if K > 1:
						break
					else: # Make sure two haplotype clusters are generated
						print(", No diversity (K=1)! Adding extra cluster.")
						cluster_cy.genClust(X, R, C, z_vec, c_vec, n_vec, u_vec, U, K)
						K += 1
			else:
				memoryview(z_tmp)[:] = memoryview(z_vec)

			# Count sizes and construct marginal medians
			cluster_cy.marginalMedians(R, C, n_vec, K)

		# Iterative re-clustering of haplotypes
		if K > 2:
			# Set singleton clusters to zero from right hand side
			cluster_cy.setZero(n_vec, N_mac-1, K)
			cluster_cy.marginalMedians(R, C, n_vec, K)
			cluster_cy.assignClust(X, R, C, z_vec, c_vec, n_vec, n_tmp, u_vec, U, K)
			cluster_cy.updateN(n_vec, n_tmp, K)
			K_tmp = np.sum(n_vec > 0, dtype=np.uint32)

			# Remove smallest clusters iterativly
			while K_tmp > 2:
				# Re-assign haplotypes
				cluster_cy.marginalMedians(R, C, n_vec, K)
				cluster_cy.assignClust(X, R, C, z_vec, c_vec, n_vec, n_tmp, u_vec, U, K)
				cluster_cy.updateN(n_vec, n_tmp, K)

				# Find smallest cluster
				N_min = cluster_cy.findZero(n_vec, N, N_mac, K)
				if N_min >= N_mac: # Ensure convergence
					if cluster_cy.countDist(z_vec, z_tmp, U) == 0:
						break
				else:
					K_tmp -= 1 # Cluster removed
					memoryview(z_tmp)[:] = memoryview(z_vec)

			# Re-cluster K = 2 case
			if (K_tmp == 2) and (N_min < N_mac):
				cluster_cy.assignClust(X, R, C, z_vec, c_vec, n_vec, n_tmp, u_vec, U, K)
				cluster_cy.updateN(n_vec, n_tmp, K)

		# Fix cluster median and cluster assignment order
		cluster_cy.medianFix(R, z_vec, n_vec, K, U)
		cluster_cy.assignFix(Z, z_vec, p_vec, d_vec, np.uint32(w))
		K = np.sum(n_vec > 0, dtype=np.uint32)
		k_vec[w] = K

		# Generate optional saves (medians)
		if args.medians:
			cluster_cy.estimateLoglike(R, C, L, n_vec, K)
			N_arr = np.append(N_arr, n_vec[:K])
			with open(f"{args.out}.bcm", "ab") as f:
				R[:K].tofile(f)
			with open(f"{args.out}.blk", "ab") as f:
				L[:K,:K].tofile(f)

		# Reset arrays
		cluster_cy.resetArrays(c_vec, n_vec, p_vec, d_vec, u_vec)

	# Release memory
	del G, X, C, z_vec, c_vec, z_tmp, p_vec, d_vec, n_vec, u_vec, a_tmp, b_tmp, d_tmp, e_tmp, n_tmp
	if args.memory:
		del H
	print(".\n")

	# Extract window information
	s_vec = v_vec[w_vec[:-1]].copy()
	if args.size is not None:
		e_vec = v_vec[w_vec[:-1] + args.size-1].copy()
		e_vec[-1] = v_vec[-1]
		b_vec = np.full(W, args.size, dtype=np.uint32)
		b_vec[-1] = w_vec[-1] - w_vec[-2]
	else:
		e_vec = v_vec[w_vec[1:] - 1]
		b_vec = w_vec[1:] - w_vec[:-1]
	del v_vec, w_vec

	# Create window information array
	w_mat = np.hstack((
		np.array([chrom]).repeat(W).reshape(-1, 1),
		s_vec.reshape(-1, 1), e_vec.reshape(-1, 1), 
		(e_vec - s_vec).reshape(-1, 1),
		b_vec.reshape(-1, 1), k_vec.reshape(-1, 1)
	))

	# Save hapla output and print info
	h_win = ["#CHROM", "START", "END", "LENGTH", "SIZE", "K"]
	with open(f"{args.out}.bca", "wb") as f:
		np.array([7, 9, 13], dtype=np.uint8).tofile(f) # Add magic numbers
		Z.tofile(f) # Save haplotype cluster assignments to binary file format
	np.savetxt(f"{args.out}.ids", s_list, fmt="%s")
	np.savetxt(f"{args.out}.win", w_mat, fmt="%s", delimiter="\t", comments="", header="\t".join(h_win))
	print("\rSaved haplotype clusters in binary format:\n" + \
		f"- {args.out}.bca\n" + \
		f"- {args.out}.ids\n" + \
		f"- {args.out}.win\n")
	if args.medians: # Save haplotype cluster medians to binary file format
		np.savetxt(f"{args.out}.hcc", N_arr, fmt="%i")
		print(f"Saved haplotype cluster medians in binary format:\n" + \
			f"- {args.out}.bcm\n" + \
			f"- {args.out}.blk\n" + \
			f"- {args.out}.wix\n" + \
			f"- {args.out}.hcc\n")
		del N_arr
	del e_vec, w_mat

	# Save haplotype cluster assignments in binary PLINK format
	if args.plink:
		print("\rGenerating binary PLINK output.", end="")
		K_tot = np.sum(k_vec, dtype=np.uint32)
		P_mat = np.zeros((K_tot, 3), dtype=np.uint32)
		Z_bin = np.zeros((K_tot, B), dtype=np.uint8)
		c_vec = np.insert(np.cumsum(k_vec[:-1], dtype=np.uint32), 0, 0)
		reader_cy.convertPlink(Z, Z_bin, P_mat, k_vec, c_vec, b_vec)
		
		# Save .bed file including magic numbers
		with open(f"{args.out}.bed", "w") as bfile:
			np.array([108, 27, 1], dtype=np.uint8).tofile(bfile)
			Z_bin.tofile(bfile)
		del b_vec, c_vec, Z_bin, Z

		# Save .bim file
		tmp = np.array([f"{chrom}_W{w}_K{k}_B{l}" for w,k,l in P_mat])
		bim = np.hstack((
			np.array([chrom]).repeat(K_tot).reshape(-1, 1),
			tmp.reshape(-1, 1),
			np.zeros((K_tot, 1), dtype=np.uint32),
			s_vec.repeat(k_vec).reshape(-1, 1),
			np.array(["K"]).repeat(K_tot).reshape(-1, 1),
			np.zeros((K_tot, 1), dtype=np.uint32)
		))
		np.savetxt(f"{args.out}.bim", bim, fmt="%s", delimiter="\t")
		del k_vec, s_vec, bim, tmp, P_mat

		# Save .fam file
		if args.duplicate_fid:
			s_list = s_list.repeat(2, axis=1)
		else:
			s_list = np.hstack((np.zeros((N//2, 1), dtype=np.uint8), s_list))
		fam = np.hstack((
			s_list,
			np.zeros((N//2, 3), dtype=np.uint8),
			np.full((N//2, 1), -9, dtype=np.int8)
		))
		np.savetxt(f"{args.out}.fam", fam, fmt="%s", delimiter="\t")

		# Print info
		print("\rSaved haplotype clusters in binary PLINK format:\n" + \
			f"- {args.out}.bed\n" + \
			f"- {args.out}.bim\n" + \
			f"- {args.out}.fam\n")
		del fam, s_list

	# Print elapsed time for parsing and total computation
	t_min = int(t_par//60)
	t_sec = int(t_par - t_min*60)
	print(f"Total parsing time: {t_min}m{t_sec}s")
	t_tot = time() - start
	t_min = int(t_tot//60)
	t_sec = int(t_tot - t_min*60)
	print(f"Total elapsed time: {t_min}m{t_sec}s")



##### Main exception #####
assert __name__ != "__main__", "Please use the 'hapla cluster' command!"



# cython: language_level=3, boundscheck=False, wraparound=False, initializedcheck=False, cdivision=True
cimport numpy as np
from cython.parallel import prange
from libc.math cimport log, exp
from libc.stdint cimport uint8_t, uint32_t

##### hapla - local ancestry inference #####
# Safe log-sum-exp for array
cdef inline double _logsumexp(
		const double* vec, const size_t K
	) noexcept nogil:
	cdef:
		double max_v = vec[0]
		double sum_v = 0.0
		size_t k
	for k in range(1, K):
		if vec[k] > max_v:
			max_v = vec[k]
	for k in range(K):
		sum_v += exp(vec[k] - max_v)
	return log(sum_v) + max_v

# Calculate emission probabilities
cpdef void calcEmissions(
		double[:,:,::1] E, const uint8_t[:,::1] Z, const double[::1] P, const uint8_t[::1] k_vec, 
		const uint32_t[::1] c_vec
	) noexcept nogil:
	cdef:
		size_t N = E.shape[0]
		size_t W = E.shape[1]
		size_t K = E.shape[2]
		size_t i, k, s, w, z, B
	for i in prange(N):
		for w in range(W):
			s = c_vec[w]
			B = k_vec[w]
			z = Z[i,w]
			for k in range(K):
				E[i,w,k] = log(P[s + B*k + z])

# Calculate transition probabilities - T[k1, k2] = P(Z_{w} = k1 | Z_{w-1} = k2)
cpdef void calcTransition(
		double[:,::1] T, const double[::1] Q, const double a
	) noexcept nogil:
	cdef:
		uint32_t K = T.shape[0]
		double e = exp(-a)
		size_t k1, k2
	for k1 in range(K):
		for k2 in range(K):
			if k1 == k2:
				T[k1,k2] = log((1.0 - e)*Q[k1] + e)
			else:
				T[k1,k2] = log((1.0 - e)*Q[k1])

# Viterbi algorithm
cpdef void viterbi(
		const double[:,::1] E, const double[::1] Q_log, const double[:,::1] T, double[:,::1] A, uint8_t[:,::1] I, 
		uint8_t[::1] V
	) noexcept nogil:
	cdef:
		double max_v
		size_t W = E.shape[0]
		size_t K = E.shape[1]
		size_t k, w, k1, k2
	# Basis step
	A[0,0] = E[0,0] + Q_log[0]
	for k in range(1, K):
		A[0,k] = E[0,k] + Q_log[k]
	
	# Loop through sequence
	for w in range(1, W):
		for k1 in range(K):
			A[w,k1] = A[w-1,0] + T[k1,0] + E[w,k1]
			I[w,k1] = 0
			for k2 in range(1, K):
				max_v = A[w - 1,k2] + T[k1,k2] + E[w,k1]
				if max_v > A[w,k1]:
					A[w,k1] = max_v
					I[w,k1] = k2
	
	# Decode path
	V[W - 1] = 0
	max_v = A[W - 1,0]
	for k in range(1, K):
		if A[W - 1,k] > max_v:
			V[W - 1] = k
			max_v = A[W - 1,k]
	for w in range(W-2, -1, -1):
		V[w] = I[w + 1,V[w + 1]]

# Forward-backward algorithm
cpdef void calcFwdBwd(
		const double[:,::1] E, double[:,::1] L, const double[::1] Q_log, const double[:,::1] T, double[:,::1] A, \
		double[:,::1] B, double[::1] v
	) noexcept nogil:
	cdef:
		double l_fwd
		size_t W = E.shape[0]
		size_t K = E.shape[1]
		size_t k, w, k1, k2
	# Forward calculations
	for k in range(K):
		A[0,k] = E[0,k] + Q_log[k]
	for w in range(1, W):
		for k1 in range(K):
			for k2 in range(K):
				v[k2] = A[w - 1,k2] + T[k1,k2]
			A[w,k1] = _logsumexp(&v[0], K) + E[w,k1]
	
	# Forward log-likelihood
	for k in range(K):
		v[k] = A[W - 1,k]
	l_fwd = _logsumexp(&v[0], K)

	# Backward calculations
	for k in range(K):
		B[W - 1,k] = 0.0
	for w in range(W-2, -1, -1):
		for k1 in range(K):
			for k2 in range(K):
				v[k2] = B[w + 1,k2] + E[w + 1,k2] + T[k2,k1]
			B[w,k1] = _logsumexp(&v[0], K)

	# Compute posterior probabilities
	for w in range(W):
		for k in range(K):
			L[w,k] = exp(A[w,k] + B[w,k] - l_fwd)



"""
Main caller of hapla.
"""

__author__ = "Jonas Meisner"

# Libraries
import argparse
import sys

# Main function
def main():
	# Argparser
	parser = argparse.ArgumentParser(prog="hapla")
	subparsers = parser.add_subparsers(title="hapla commands")

	### Commands
	# hapla cluster
	parser_cluster = subparsers.add_parser("cluster")
	parser_cluster.add_argument("-g", "--vcf", "--bcf", metavar="FILE",
		help="Input phased genotype file in VCF/BCF format")
	parser_cluster.add_argument("-f", "--size", type=int,
		metavar="INT", help="Use fixed window size")
	parser_cluster.add_argument("-w", "--windows", metavar="FILE",
		help="Use provided window indices")
	parser_cluster.add_argument("-l", "--lmbda", type=float, default=0.125,
		metavar="FLOAT", help="Set lambda hyperparameter (0.125)")
	parser_cluster.add_argument("-s", "--step", type=int,
		metavar="INT", help="Step-size for sliding window")
	parser_cluster.add_argument("-t", "--threads", type=int, default=1,
		metavar="INT", help="Number of threads (1)")
	parser_cluster.add_argument("-o", "--out", default="hapla.cluster",
		metavar="OUTPUT", help="Output prefix")
	parser_cluster.add_argument("--min-freq", type=float, default=0.01,
		metavar="FLOAT", help="Minimum haplotype cluster frequency (0.01)")
	parser_cluster.add_argument("--min-mac", type=int,
		metavar="INT", help="Minimum haplotype cluster allele count")
	parser_cluster.add_argument("--max-clusters", type=int, default=128,
		metavar="INT", help="Maximum number of haplotype clusters per window (128)")
	parser_cluster.add_argument("--max-iterations", type=int, default=500,
		metavar="INT", help="Maximum number of iterations (500)")
	parser_cluster.add_argument("--medians", action="store_true",
		help="Save haplotype cluster medians")
	parser_cluster.add_argument("--plink", action="store_true",
		help="Generate binary PLINK output")
	parser_cluster.add_argument("--duplicate-fid", action="store_true",
		help="Use sample list as family ID (PLINK 1.9 compatibility)")
	parser_cluster.add_argument("--memory", action="store_true",
		help="Store haplotypes in 1-bit matrix")

	# hapla struct
	parser_struct = subparsers.add_parser("struct")
	parser_struct.add_argument("-f", "--filelist", metavar="FILE",
		help="Filelist with paths to haplotype cluster alleles files")
	parser_struct.add_argument("-z", "--clusters", metavar="FILE",
		help="Path to a single haplotype cluster alleles file")
	parser_struct.add_argument("-t", "--threads", type=int, default=1,
		metavar="INT", help="Number of threads (1)")
	parser_struct.add_argument("-o", "--out", default="hapla.pca",
		metavar="OUTPUT", help="Output prefix")
	parser_struct.add_argument("--grm", action="store_true",
		help="Estimate genome-wide relationship matrix (GRM)")
	parser_struct.add_argument("--batch", type=int, default=8192,
		metavar="INT", help="Number of cluster alleles in batches (8192)")
	parser_struct.add_argument("--power", type=int, default=11,
		metavar="INT", help="Number of power iterations to perform (11)")
	parser_struct.add_argument("--no-centering", action="store_true",
		help="Do not perform Gower and data centering of GRM")
	parser_struct.add_argument("--duplicate-fid", action="store_true",
		help="Use sample list as family ID for GCTA format")
	parser_struct.add_argument("--pca", type=int,
		metavar="INT", help="Perform PCA and extract eigenvectors")
	parser_struct.add_argument("--seed", type=int, default=42,
		metavar="INT", help="Random seed (42)")
	parser_struct.add_argument("--loadings", action="store_true",
		help="Save loadings of SVD")
	parser_struct.add_argument("--raw", action="store_true",
		help="Raw output without '*.fam' info")

	# hapla predict
	parser_predict = subparsers.add_parser("predict")
	parser_predict.add_argument("-g", "--vcf", "--bcf", metavar="FILE",
		help="Input phased genotype file in VCF/BCF format")
	parser_predict.add_argument("-r", "--ref", metavar="FILE",
		help="Input reference prefix of pre-estimated cluster medians")
	parser_predict.add_argument("-t", "--threads", type=int, default=1,
		metavar="INT", help="Number of threads (1)")
	parser_predict.add_argument("-o", "--out", default="hapla.predict",
		metavar="OUTPUT", help="Output prefix")
	parser_predict.add_argument("--plink", action="store_true",
		help="Generate binary PLINK output")
	parser_predict.add_argument("--duplicate-fid", action="store_true",
		help="Use sample list as family ID (PLINK 1.9 compatibility)")
	parser_predict.add_argument("--memory", action="store_true",
		help="Store haplotypes in 2-bit matrix")

	# hapla admix
	parser_admix = subparsers.add_parser("admix")
	parser_admix.add_argument("-f", "--filelist", metavar="FILE",
		help="Filelist with paths to haplotype cluster alleles files")
	parser_admix.add_argument("-z", "--clusters", metavar="FILE",
		help="Path to a single haplotype cluster alleles file")
	parser_admix.add_argument("-k", "--K", type=int,
		metavar="INT", help="Number of ancestral components")
	parser_admix.add_argument("-t", "--threads", type=int, default=1,
		metavar="INT", help="Number of threads (1)")
	parser_admix.add_argument("-o", "--out", default="hapla.admix",
		metavar="OUTPUT", help="Output prefix")
	parser_admix.add_argument("--seed", type=int, default=42,
		metavar="INT", help="Random seed (42)")
	parser_admix.add_argument("--iter", type=int, default=1000,
		metavar="INT", help="Maximum number of iterations (1000)")
	parser_admix.add_argument("--tole", type=float, default=1.0,
		metavar="FLOAT", help="Tolerance in log-likelihood between iterations (1.0)")
	parser_admix.add_argument("--check", type=int, default=10,
		metavar="INT", help="Check for convergence every c-th iteration (10)")
	parser_admix.add_argument("--supervised", metavar="FILE",
		help="Path to population assignment file")
	parser_admix.add_argument("--no-freq", action="store_true",
		help="Do not save haplotype cluster frequencies")

	# hapla fatash
	parser_fatash = subparsers.add_parser("fatash")
	parser_fatash.add_argument("-f", "--filelist", metavar="FILE",
		help="Filelist with paths to haplotype cluster alleles files")
	parser_fatash.add_argument("-z", "--clusters", metavar="FILE",
		help="Path to a single haplotype cluster alleles file")
	parser_fatash.add_argument("-e", "--pfilelist", metavar="FILE",
		help="Filelist with paths to haplotype cluster frequencies files")
	parser_fatash.add_argument("-p", "--pfile", metavar="FILE",
		help="Path to file with haplotype cluster frequencies")
	parser_fatash.add_argument("-q", "--qfile", metavar="FILE",
		help="Path to file with admixture proportions")
	parser_fatash.add_argument("-t", "--threads", type=int, default=1,
		metavar="INT", help="Number of threads (1)")
	parser_fatash.add_argument("-o", "--out", default="hapla.fatash",
		metavar="OUTPUT", help="Output prefix")
	parser_fatash.add_argument("--alpha", type=float, default=1e-8,
		metavar="FLOAT", help="Set fixed alpha rate (1e-8)")
	parser_fatash.add_argument("--viterbi", action="store_true",
		help="Perform Viterbi decoding")
	parser_fatash.add_argument("--save-posterior", action="store_true",
		help="Save posterior probabilities")

	# Parse arguments
	args = parser.parse_args()
	if len(sys.argv) < 2:
		parser.print_help()
		sys.exit()

	### Run specified command
	# hapla cluster
	if sys.argv[1] == "cluster":
		if len(sys.argv) < 3:
			parser_cluster.print_help()
			sys.exit()
		else:
			from hapla import cluster
			cluster.main(args)
	
	# hapla struct
	if sys.argv[1] == "struct":
		if len(sys.argv) < 3:
			parser_struct.print_help()
			sys.exit()
		else:
			from hapla import struct
			struct.main(args)

	# hapla predict
	if sys.argv[1] == "predict":
		if len(sys.argv) < 3:
			parser_predict.print_help()
			sys.exit()
		else:
			from hapla import predict
			predict.main(args)

	# hapla admix
	if sys.argv[1] == "admix":
		if len(sys.argv) < 3:
			parser_admix.print_help()
			sys.exit()
		else:
			from hapla import admix
			admix.main(args)

	# hapla fatash
	if sys.argv[1] == "fatash":
		if len(sys.argv) < 3:
			parser_fatash.print_help()
			sys.exit()
		else:
			from hapla import fatash
			fatash.main(args)



##### Define main #####
if __name__ == "__main__":
	main()



"""
hapla.
Haplotype clustering using pre-estimated cluster medians.
"""

__author__ = "Jonas Meisner"

# Libraries
import os
from time import time

##### hapla predict #####
def main(args):
	print("-----------------------------------")
	print("hapla by Jonas Meisner (v0.23.0)")
	print(f"hapla predict using {args.threads} thread(s)")
	print("-----------------------------------\n")
	
	# Check input
	assert args.vcf is not None, "Please provide phased genotype file (--bcf or --vcf)!"
	assert os.path.isfile(f"{args.vcf}"), "VCF/BCF file doesn't exist!"
	assert os.path.isfile(f"{args.vcf}.csi") or os.path.isfile(f"{args.vcf}.tbi"), "VCF/BCF index doesn't exist!"
	assert args.ref is not None, "Please provide pre-estimated reference haplotype cluster medians (--ref)!"
	assert os.path.isfile(f"{args.ref}.bcm"), "bcm file doesn't exist!"
	assert os.path.isfile(f"{args.ref}.win"), "win file doesn't exist!"
	assert os.path.isfile(f"{args.ref}.wix"), "wix file doesn't exist!"
	assert os.path.isfile(f"{args.ref}.hcc"), "hcc file doesn't exist!"
	assert args.threads > 0, "Please select a valid number of threads!"
	start = time()

	# Control threads of external numerical libraries
	os.environ["MKL_NUM_THREADS"] = str(args.threads)
	os.environ["MKL_MAX_THREADS"] = str(args.threads)
	os.environ["OMP_NUM_THREADS"] = str(args.threads)
	os.environ["OMP_MAX_THREADS"] = str(args.threads)
	os.environ["NUMEXPR_NUM_THREADS"] = str(args.threads)
	os.environ["NUMEXPR_MAX_THREADS"] = str(args.threads)
	os.environ["OPENBLAS_NUM_THREADS"] = str(args.threads)
	os.environ["OPENBLAS_MAX_THREADS"] = str(args.threads)

	# Import numerical libraries and cython functions
	import numpy as np
	from cyvcf2 import VCF
	from math import ceil
	from hapla import reader_cy
	from hapla import memory_cy
	from hapla import shared_cy

	# Initiate VCF and extract sample list
	print("\rLoading VCF/BCF file...", end="")
	v_file = VCF(args.vcf, threads=min(args.threads, 4))
	s_list = np.array(v_file.samples).reshape(-1, 1)
	N = 2*s_list.shape[0]
	M = v_file.num_records
	B = ceil(N/4)

	# Allocate arrays
	if args.memory:
		G = np.zeros((M, B), dtype=np.uint8)
	else:
		G = np.zeros((M, N), dtype=np.uint8)
	v_vec = np.zeros(M, dtype=np.uint32)

	# Read variants into matrix
	v_file = VCF(args.vcf, threads=min(args.threads, 4))
	for j, variant in enumerate(v_file):
		V = variant.genotype.array()
		if args.memory:
			memory_cy.predBit(G[j], V, N//2)
		else:
			reader_cy.predVar(G[j], V, N//2)
		v_vec[j] = variant.POS
	chrom = str(variant.CHROM) # Extract chromosome information
	del V, v_file
	print(f"\rLoaded phased genotype data: {N} haplotypes and {M} SNPs.")

	# Load window information from reference
	w_mat = np.genfromtxt(f"{args.ref}.win", dtype=np.str_, skip_header=1)
	assert w_mat[0,0] == chrom, "Chromosome name differ between files!"
	assert int(w_mat[0,1]) == v_vec[0], "Positions differ between files!"
	assert int(w_mat[-1,2]) == v_vec[-1], "Positions differ between files!"
	s_vec = w_mat[:,1].astype(np.uint32)
	b_vec = w_mat[:,4].astype(np.uint32)
	k_vec = w_mat[:,5].astype(np.uint8)
	W = k_vec.shape[0]

	# Haplotype cluster medians
	with open(f"{args.ref}.bcm", "rb") as f:
		# Check magic numbers
		m_vec = np.fromfile(f, dtype=np.uint8, count=3)
		assert np.allclose(m_vec, np.array([7, 9, 13], dtype=np.uint8)), "Magic number doesn't match file format!"
		R_arr = np.fromfile(f, dtype=np.uint8)
		
	# Load window setup files
	w_vec = np.genfromtxt(f"{args.ref}.wix", dtype=np.uint32)
	N_arr = np.genfromtxt(f"{args.ref}.hcc", dtype=np.uint32)
	assert np.allclose(v_vec[w_vec], s_vec), "SNP set doesn't match!"
	del v_vec

	# Containers
	Z = np.zeros((W, N), dtype=np.uint8) # Haplotype cluster alleles

	# Clustering
	B = 0
	K = 0
	print(f"Clustering {W} windows.")
	for w in np.arange(W):
		print(f"\rWindow {w+1}/{W}", end="")
		b_win = int(b_vec[w])
		k_win = int(k_vec[w])

		# Load haplotype window
		R_mat = R_arr[B:(B + k_win*b_win)]
		R_mat.shape = (k_win, b_win)
		n_vec = N_arr[K:(K + k_win)]
		X = np.zeros((N, R_mat.shape[1]), dtype=np.uint8)
		if args.memory:
			memory_cy.predictBit(G, X, w_vec[w])
		else:
			reader_cy.predictHap(G, X, w_vec[w])
		
		# Cluster assignment
		shared_cy.predictCluster(X, R_mat, Z, n_vec, k_vec[w], w)

		# Update counter
		B += k_win*b_win
		K += k_win
	del G, X, R_mat, R_arr, N_arr, n_vec, w_vec
	print(".\n")

	# Save hapla output and print info
	h_win = ["#CHROM", "START", "END", "LENGTH", "SIZE", "K"]
	with open(f"{args.out}.bca", "wb") as f:
		np.array([7, 9, 13], dtype=np.uint8).tofile(f) # Add magic numbers
		Z.tofile(f) # Save haplotype cluster assignments to binary file
	np.savetxt(f"{args.out}.ids", s_list, fmt="%s")
	np.savetxt(f"{args.out}.win", w_mat, fmt="%s", delimiter="\t", comments="", header="\t".join(h_win))
	print("\rSaved haplotype clusters in binary hapla format:\n" + \
		f"- {args.out}.bca\n" + \
		f"- {args.out}.ids\n" + \
		f"- {args.out}.win\n")

	# Save haplotype cluster assignments in binary PLINK format
	if args.plink:
		print("\rGenerating binary PLINK output.", end="")
		B = ceil(N/8)
		K_tot = np.sum(k_vec, dtype=int)
		P_mat = np.zeros((K_tot, 3), dtype=np.uint32)
		Z_bin = np.zeros((K_tot, B), dtype=np.uint8)
		c_vec = np.insert(np.cumsum(k_vec[:-1], dtype=np.uint32), 0, 0)
		reader_cy.convertPlink(Z, Z_bin, P_mat, k_vec, c_vec, b_vec)
		
		# Save .bed file including magic numbers
		with open(f"{args.out}.bed", "w") as bfile:
			np.array([108, 27, 1], dtype=np.uint8).tofile(bfile)
			Z_bin.tofile(bfile)
		del c_vec, b_vec, Z_bin, Z

		# Save .bim file
		tmp = np.array([f"{chrom}_W{w}_K{k}_B{l}" for w,k,l in P_mat])
		bim = np.hstack((
			np.array([chrom]).repeat(K_tot).reshape(-1, 1),
			tmp.reshape(-1, 1),
			np.zeros((K_tot, 1), dtype=np.uint32),
			s_vec.repeat(k_vec).reshape(-1, 1),
			np.array(["K"]).repeat(K_tot).reshape(-1, 1),
			np.zeros((K_tot, 1), dtype=np.uint32)
		))
		np.savetxt(f"{args.out}.bim", bim, fmt="%s", delimiter="\t")
		del k_vec, s_vec, bim, tmp, P_mat
		
		# Save .fam file
		if args.duplicate_fid:
			s_list = s_list.repeat(2, axis=1)
		else:
			s_list = np.hstack((np.zeros((N//2, 1), dtype=np.uint8), s_list))
		fam = np.hstack((
			s_list,
			np.zeros((N//2, 3), dtype=np.uint8),
			np.full((N//2, 1), -9, dtype=np.int8)
		))
		np.savetxt(f"{args.out}.fam", fam, fmt="%s", delimiter="\t")

		# Print info
		print("\rSaved haplotype cluster alleles in binary PLINK format:\n" + \
			f"- {args.out}.bed\n" + \
			f"- {args.out}.bim\n" + \
			f"- {args.out}.fam\n")
		del fam, s_list

	# Print elapsed time for computation
	t_tot = time() - start
	t_min = int(t_tot//60)
	t_sec = int(t_tot - t_min*60)
	print(f"Total elapsed time: {t_min}m{t_sec}s")



##### Main exception #####
assert __name__ != "__main__", "Please use the 'hapla predict' command!"



from setuptools import setup, Extension
from Cython.Build import cythonize
import numpy

extensions = [
	Extension(
		"hapla.reader_cy",
		["hapla/reader_cy.pyx"],
		extra_compile_args=['-fopenmp', '-O3', '-ffast-math', '-march=native'],
		extra_link_args=['-fopenmp', '-lm'],
		include_dirs=[numpy.get_include()],
		define_macros=[('NPY_NO_DEPRECATED_API', 'NPY_1_7_API_VERSION')]
	),
	Extension(
		"hapla.cluster_cy",
		["hapla/cluster_cy.pyx"],
		extra_compile_args=['-fopenmp', '-O3', '-ffast-math', '-march=native'],
		extra_link_args=['-fopenmp', '-lm'],
		include_dirs=[numpy.get_include()],
		define_macros=[('NPY_NO_DEPRECATED_API', 'NPY_1_7_API_VERSION')]
	),
	Extension(
		"hapla.shared_cy",
		["hapla/shared_cy.pyx"],
		extra_compile_args=['-fopenmp', '-O3', '-ffast-math', '-march=native'],
		extra_link_args=['-fopenmp', '-lm'],
		include_dirs=[numpy.get_include()],
		define_macros=[('NPY_NO_DEPRECATED_API', 'NPY_1_7_API_VERSION')]
	),
	Extension(
		"hapla.memory_cy",
		["hapla/memory_cy.pyx"],
		extra_compile_args=['-fopenmp', '-O3', '-ffast-math', '-march=native'],
		extra_link_args=['-fopenmp', '-lm'],
		include_dirs=[numpy.get_include()],
		define_macros=[('NPY_NO_DEPRECATED_API', 'NPY_1_7_API_VERSION')]
	),
	Extension(
		"hapla.admix_cy",
		["hapla/admix_cy.pyx"],
		extra_compile_args=['-fopenmp', '-O3', '-ffast-math', '-march=native'],
		extra_link_args=['-fopenmp', '-lm'],
		include_dirs=[numpy.get_include()],
		define_macros=[('NPY_NO_DEPRECATED_API', 'NPY_1_7_API_VERSION')]
	),
	Extension(
		"hapla.fatash_cy",
		["hapla/fatash_cy.pyx"],
		extra_compile_args=['-fopenmp', '-O3', '-ffast-math', '-march=native'],
		extra_link_args=['-fopenmp', '-lm'],
		include_dirs=[numpy.get_include()],
		define_macros=[('NPY_NO_DEPRECATED_API', 'NPY_1_7_API_VERSION')]
	)
]

setup(
	name="hapla",
	version="0.23.0",
	author="Jonas Meisner",
	author_email="meisnerucph@gmail.com",
	description="Framework for haplotype clustering in phased genotype data",
	long_description_content_type="text/markdown",
	long_description=open("README.md").read(),
	url="https://github.com/Rosemeis/hapla",
	classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta",
    ],
	ext_modules=cythonize(extensions, compiler_directives={'language_level':'3'}),
	python_requires=">=3.10",
	install_requires=[
		"cython>3.0.0",
		"cyvcf2>=0.31.0",
		"numpy>2.0.0"
	],
	packages=["hapla"],
	entry_points={
		"console_scripts": ["hapla=hapla.main:main"]
	},
)



# cython: language_level=3, boundscheck=False, wraparound=False, initializedcheck=False, cdivision=True
cimport numpy as np
from libc.stdint cimport uint8_t, int16_t, uint32_t

##### Cython memory efficient functions #####
# Read variant from VCF/BCF into 1-bit integer format
cpdef void readBit(
		uint8_t[::1] G, const int16_t[:,::1] V, const size_t N
	) noexcept nogil:
	cdef:
		size_t B = G.shape[0]
		size_t i = 0
		size_t b, bit
	for b in range(B):
		for bit in range(0, 8, 2):
			if V[i,0] == 1:
				G[b] |= (1<<bit)
			if V[i,1] == 1:
				G[b] |= (1<<(bit+1))
			
			# Increase counter and check for break
			i += 1
			if i == N:
				break

# Read variant from VCF/BCF into 2-bit integer format
cpdef void predBit(
		uint8_t[::1] G, const int16_t[:,::1] V, const size_t N
	) noexcept nogil:
	cdef:
		size_t B = G.shape[0]
		size_t i = 0
		size_t b, bit
	for b in range(B):
		for bit in range(0, 8, 4):
			if V[i,0] == 1: # Allele 1 (1,1)
				G[b] |= (1<<bit)
				G[b] |= (1<<(bit+1))
			elif V[i,0] == -1: # Missing (1,0)
				G[b] |= (1<<bit)
			if V[i,1] == 1: # Allele 2 (1,1)
				G[b] |= (1<<(bit+2))
				G[b] |= (1<<(bit+3))
			elif V[i,1] == -1: # Missing (1,0)
				G[b] |= (1<<(bit+2))
			
			# Increase counter and check for break
			i += 1
			if i == N:
				break

# Convert 1-bit into full array and initialize cluster mean
cpdef void convertBit(
		const uint8_t[:,::1] G, uint8_t[:,::1] H, uint32_t[:,::1] C, uint32_t[::1] p_vec, uint32_t[::1] d_vec, \
		uint32_t[::1] a_tmp, uint32_t[::1] b_tmp, uint32_t[::1] d_tmp, uint32_t[::1] e_tmp, const size_t S
	) noexcept nogil:
	cdef:
		size_t B = G.shape[1]
		size_t M = H.shape[0]
		size_t N = H.shape[1]
		size_t b, h, i, j, k, s, u, v, bit
		uint8_t mask = 1
		uint8_t g, byte
		uint32_t f, l, p, q
	# Populate haplotype matrix and cluster mean
	for j in range(M):
		h = 0
		s = S + j
		C[0,j] = 0
		for b in range(B):
			byte = G[s,b]
			for bit in range(8):
				g = (byte >> bit) & mask
				H[j,h] = g
				C[0,j] += g
				h += 1
				if h == N:
					break
	
		# Populate suffix arrays
		u = v = 0
		p = q = <uint32_t>j + 1
		for i in range(N):
			f = p_vec[i]
			l = d_vec[i]
			if l > p:
				p = l
			if l > q:
				q = l
			if H[j,f] == 0:
				a_tmp[u] = f
				d_tmp[u] = p
				u += 1
				p = 0
			else:
				b_tmp[v] = f
				e_tmp[v] = q
				v += 1
				q = 0
		for k in range(u):
			p_vec[k] = a_tmp[k]
			d_vec[k] = d_tmp[k]
		for k in range(v):
			p_vec[u+k] = b_tmp[k]
			d_vec[u+k] = e_tmp[k]

# Extract unique haplotypes from suffix arrays
cpdef uint32_t uniqueBit(
		const uint8_t[:,::1] H, uint8_t[:,::1] X, const uint32_t[::1] p_vec, const uint32_t[::1] d_vec, 
		uint32_t[::1] u_vec
	) noexcept nogil:
	cdef:
		size_t N = X.shape[0]
		size_t M = X.shape[1]
		size_t u = 0
		size_t h, i, j
	for i in range(N):
		if d_vec[i] != 0:
			h = p_vec[i]
			for j in range(M):
				X[u,j] = H[j,h]
			u += 1
		u_vec[u-1] += 1
	return u

# Convert 2-bit into full array for predicting target clusters
cpdef void predictBit(
		const uint8_t[:,::1] G, uint8_t[:,::1] X, const size_t m
	) noexcept nogil:
	cdef:
		uint8_t[4] recode = [0, 9, 9, 1]
		uint8_t mask = 3
		uint8_t i, s, byte
		size_t B = G.shape[1]
		size_t N = X.shape[0]
		size_t M = X.shape[1]
		size_t b, j, bit
	for j in range(M):
		i = 0
		s = m + j
		for b in range(B):
			byte = G[s,b]
			for bit in range(4):
				X[i,j] = recode[byte & mask]
				byte = byte >> 2 # Right shift 2 bits
				i += 1
				if i == N:
					break



# cython: language_level=3, boundscheck=False, wraparound=False, initializedcheck=False, cdivision=True
cimport numpy as np
from cython.parallel import prange
from libc.stdint cimport uint8_t, uint32_t

##### hapla - analyses on haplotype cluster assignments #####
### hapla struct
# Extract aggregated haplotype cluster counts
cpdef void haplotypeAggregate(
		const uint8_t[:,::1] Z, uint8_t[:,::1] Z_agg, double[::1] p, const uint8_t[::1] k_vec, 
		const uint32_t[::1] c_vec
	) noexcept nogil:
	cdef:
		size_t W = Z.shape[0]
		size_t N = Z.shape[1]
		size_t c, i, l, s, w
		double d = 1.0/<double>N
	for w in prange(W):
		s = <size_t>c_vec[w]
		for c in range(k_vec[w]):
			l = s + c
			for i in range(N):
				if Z[w,i] == c:
					Z_agg[l,i//2] += 1
					p[l] += 1.0
			p[l] *= d

# Center batch haplotype cluster assignment matrix
cpdef void centerZ(
		const uint8_t[:,::1] Z_agg, float[:,::1] Z_bat, const double[::1] p, const size_t m
	) noexcept nogil:
	cdef:
		float u
		size_t M = Z_bat.shape[0]
		size_t N = Z_bat.shape[1]
		size_t i, j, l
	for j in prange(M):
		l = m+j
		u = 2.0*<float>p[l]
		for i in range(N):
			Z_bat[j,i] = Z_agg[l,i] - u

# Standardize permuted batch haplotype cluster assignment matrix
cpdef void blockZ(
		const uint8_t[:,::1] Z_agg, double[:,::1] Z_bat, const double[::1] p, const double[::1] a, 
		const uint32_t[::1] s, const size_t m
	) noexcept nogil:
	cdef:
		double d, u
		size_t M = Z_bat.shape[0]
		size_t N = Z_bat.shape[1]
		size_t i, j, l
	for j in prange(M):
		l = s[m+j]
		d = a[l]
		u = 2.0*p[l]
		for i in range(N):
			Z_bat[j,i] = (Z_agg[l,i] - u)*d

# Standardize batch haplotype cluster assignment matrix
cpdef void batchZ(
		const uint8_t[:,::1] Z_agg, double[:,::1] Z_bat, const double[::1] p, const double[::1] a, const size_t m
	) noexcept nogil:
	cdef:
		double d, u
		size_t M = Z_bat.shape[0]
		size_t N = Z_bat.shape[1]
		size_t i, j, l
	for j in prange(M):
		l = m+j
		d = a[l]
		u = 2.0*p[l]
		for i in range(N):
			Z_bat[j,i] = (Z_agg[l,i] - u)*d



### hapla predict
# Calculate Hamming distance
cdef inline uint32_t hammingPred(
		const uint8_t* X, const uint8_t* R, const size_t M
	) noexcept nogil:
	cdef:
		size_t j
		uint32_t dist = 0
	for j in range(M):
		if X[j] != R[j] and X[j] != 9: # Ignore missing
			dist += 1
	return dist

# Haplotype cluster assignment based on pre-estimated medians
cpdef void predictCluster(
		uint8_t[:,::1] X, const uint8_t[:,::1] R, uint8_t[:,::1] Z, const uint32_t[::1] n_vec, const size_t K, 
		const size_t w
	) noexcept nogil:
	cdef:
		size_t c, d, i, k, z
		size_t N = X.shape[0]
		size_t M = X.shape[1]
		uint8_t* xi
	for i in prange(N):
		xi = &X[i,0]
		z = 0
		c = hammingPred(xi, &R[0,0], M)
		for k in range(1, K):
			d = hammingPred(xi, &R[k,0], M)
			if d < c or (d == c and n_vec[k] > n_vec[z]):
				z = k
				c = d
		Z[w,i] = <uint8_t>z
